/*
 * 
 * ihcp_driver.c
 * code module for Linux 2.4.x driver for PCI hardcopy boards
 * 
 * Tahoma Technology
 * (formerly Ikon Corporation)
 * 107 2nd Avenue North
 * Seattle, WA, USA 98109
 *
 * 206.728.6465
 * http://www.tahomatech.com
 * tahoma@tahomatech.com
 *
 * 
 * 5 January, 1999      start of development - initial target Linux 2.0.35 (R.H. 5.1)
 * 
 * 16 March, 1999       changed request_irq call to register node name rather than driver
 *                      name with interrupt - so each board will have a unique int name
 *                      (ihcp0, ihcp1, ... same as node name)
 * 
 * 2 Sept, 1999         re-code for 2.2
 *                      this version no longer compatible with earlier linux revs
 * 
 * 11 Nov, 1999         change the way we do plx config re-load to avoid possible
 *                      future panic with bridge chips w/short bus timeouts
 * 
 * 13 March, 2000       started working w/2.3.x
 *                      new pci_dev definition required changes in the way we get
 *                      base addresses
 *                      changed file_ops struct initialization to use named fields
 *                      changed wait_queue type and added initialization
 * 
 * 24 March, 2000       incorporated kiobuf mechanism for dma from user buffers
 *                      to replace dma copy buffer used in older versions
 *                      separated int level read from board and one read from
 *                      pci_dev (so we can use the hardware level to restore
 *                      board after en eeprom reload
 *                      incorporated modified version of Kaz Kylheku's mutex stuff
 *                      for SMP
 *                      convert all IPLX accesses to PLX accesses - doing accesses via the
 *                      ihcp space may have been allowing pre-fetch from plx registers
 *                      disable address space 0 pre-fetch
 *                      make int/cstat read in intr volatile
 *                      start DMA in two steps to avoid PLX bug
 *                      pause DMA after sleep in strategy
 * 
 *
 * 30 August, 2000	use readl and writel rather than direct de-reference (for non x86
 *			architectures) this change in ihcp_reg.h
 *			SA_SHIRQ rather than SA_INTERRUPT to allow sharing and int code
 *			interruption
 *			use time_before() when delaying to avoid jiffies wrap
 *			pci_enable_device()
 *
 * 18 June, 2001	move pci_enable_device() prior to device resource accesses to try to
 *			keep non-x86 architectures happy
 *
 *                      requires 2.3/2.4 kernel
 *
 * 6 May, 2002		uses IHCP_[ALLOC|FREE]_KIOVEC macros to deal with RH7.2 (updated)
 * 			quirks - selected by IHCP_RH72PLUS flag in ihcp_reg.h
 *
 * 			use IHCP_SG_LINK macro to do .address or .page/.offset entry
 * 			of address into scatterlist - selected by IHCP_HIGHMEM flag in ihcp_reg.h
 *
 * 			if IHCP_USE_DMA_MASK is defined, sets pci_dma_mask to indicate board's DMA
 * 			addressing capabilities - for thoroughness - 32 bit should be default
 *
 * 			add flags to enable versioning if it has been enabled in the kernel
 * 			or force versioning on, or force it off
 *
 * 			move conditional variable/mutex code to this file, for simplified
 * 			compilation and linking
 *
 *			use slab.h instead of malloc.h
 *
 * 1 August, 2002	corrected use of kiobuf maplist to generate scatterlist to allow for
 * 			highmem enabled machines (up to 4G)
 *
 * 			added kernel copy buffer for case where machine is and kernel
 * 			are highmem, but kernel does not have new scatterlist capability
 * 			in this case, we will allocate a copy buffer at init time, and
 * 			stuff the kiobuf with its pages
 *
 * 			changed max_phys to orders for allocation efficiency in case of
 * 			copy buf
 *
 * 30 March, 2005	minor type mods for x86_64 compatibility
 *
 */

/*
 * This code released under the GPL, and in the public domain
 * References to IKON left in place for compatibility and historical reasons
 *
 * Please send us code modifications and BUG reports
 */

/*
 * linux includes
 */

/*
 * versioning, no versioning, or autoversioning, depending on compile flags
 * IHCP_AUTO_VERSIONING is the default if IHCP_VERSIONING and IHCP_NO_VERSIONING
 * not defined
 * auto versioning attempts to match the versioning mode of the kernel
 */

#if	defined(IHCP_VERSIONING)
#define		MODVERSIONS
#elif	defined(IHCP_NO_VERSIONING)
#undef		MODVERSIONS
#else
#include 	<linux/config.h>
#if	defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define 	MODVERSIONS
#endif
#endif

#ifdef	MODVERSIONS
#include 	<linux/modversions.h>
#endif

#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/iobuf.h>
#include <linux/timer.h>
#include <asm/pci.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/byteorder.h>
#include <linux/types.h>
#include <asm/uaccess.h>

/*
 * ikon includes
 */

#include "ihcp_io.h"
#include "ihcp_reg.h"
#include "ihcp_var.h"

/*
 * prototype functions to keep compiler happy
 */

static void	ihcp_free_units(void);
static int	ihcp_open(struct inode *, struct file *);
static int	ihcp_close(struct inode *, struct file *);
static ssize_t	ihcp_write(struct file *, const char *, size_t, loff_t *);
static int	ihcp_ioctl(struct inode *, struct file *, u_int, u_long);
static void	ihcp_intr(int, void *, struct pt_regs *);
static int	ihcp_strategy(struct ihcp_unit_t *, const char *, int);

/*
 * mutex and conditional variable stuff - was in ikon_mutex.h (Kaz Kylheku & wdw)
 */

void mutex_init(mutex_t *);
void mutex_lock(mutex_t *);
void mutex_unlock(mutex_t *);

void cond_init(cond_t *);
int cond_timed_wait_rel(cond_t *, mutex_t *, long);
void cond_broadcast(cond_t *);

/*
 * configuration variables that may be specified at module install time
 * see ihcp_io.h for possible values
 * 
 * each internal variable is paired with one that can be modified by insmod
 */

/*
 * modifiable variables
 */

static int byte_order_def = -1;	/* default byte ordering                        */
static int vers_speed_def = -1;	/* default versatec handshake speed             */
static int cent_speed_def = -1;	/* centronics handshake speed                   */
static int mode_def = -1;	/* print/plot mode for versatec                 */
static int dma_time_def = -1;	/* dma timeout in seconds                       */
static int fifo_time_def = -1;	/* ready and fifo 1/2 full timeout in seconds   */
static int max_boards = -1;	/* max number of boards allowed                 */
static int max_phys_order = -1;	/* 2**max_phys_order * PAGE_SIZE = max xfer     */
				/* and copy buf size (if used)			*/

/*
 * declare the above as parameters if 2.1 or later -- these compile to nothing if <2.1
 * each is declared as a single integer value
 */

MODULE_PARM(byte_order_def, "1i");
MODULE_PARM(vers_speed_def, "1i");
MODULE_PARM(cent_speed_def, "1i");
MODULE_PARM(mode_def, "1i");
MODULE_PARM(dma_time_def, "1i");
MODULE_PARM(fifo_time_def, "1i");
MODULE_PARM(max_boards, "1i");
MODULE_PARM(max_phys_order, "1i");

/*
 * internal variables will use defaults if paired modifiable variables are still -1 after insmod
 * most of these variables have working versions kept in the unit structures on a per-board basis
 */

static int ihcp_byte_order_def;
static int ihcp_vers_speed_def;
static int ihcp_cent_speed_def;
static int ihcp_mode_def;
static int ihcp_dma_time_def;
static int ihcp_fifo_time_def;
static int ihcp_max_boards;
static int ihcp_max_phys_order;


/*
 * some global variables
 */

static int ihcp_boards_found;		/* number of boards actually detected           */
static int ihcp_boards_attached;	/* number of boards successfully attached       */
static int ihcp_module_initialized;	/* non zero if any board attaches and module initialized successfully   */
static int ihcp_device_major;		/* major number dynamically assigned to this driver     */

static int ihcp_max_buf_pages;		/* # of pages in copy buf = 2**ihcp_max_phys_order	*/
static int ihcp_max_buf_bytes;		/* in bytes = ihcp_max_buf_pages * PAGE_SIZE        	*/
static int ihcp_max_sg_length;		/* ihcp_max_buf_pages + 1 (allows for offset)        	*/
					/* MUST TRACK KIOBUF MAP SIZE AND COPY BUF SIZE         */
static int ihcp_iopb_mem_size;		/* size of mem allocated for iopb list                  */

/*
 * local zero in used w/alloc|free_kiovec_sz functions (if those functions used)
 */

IHCP_ZERO;

/*
 * pointer at base of unit structure array
 * array, rather than linked list, to try to be a little more efficient with kernel memory
 */

static struct ihcp_unit_t *ihcp_unit_base_p;


/*
 * declare dummy variables used to the left of = when we read a board
 * register to force the write cache to flush a register write to the bus
 */

static volatile u_long read_dump_0;
static volatile u_long read_dump_1;

/*
 * declare our file-ops structure
 * 
 * some of these have moved since 2.0
 * 
 * use named fields as of 2.3
 */

struct file_operations ihcp_fops = {
	write:ihcp_write,
	ioctl:ihcp_ioctl,
	open:ihcp_open,
	release:ihcp_close,
};





/*
 * get the global parameters, kmalloc and zero the unit structure array
 * 
 * register interrupts, map the working registers,  get the dma iopb memory & scatterlistlist, and initialize
 * things on a per-instance basis (and allocate a copy buffer, if necessary)
 * 
 * register the driver
 * 
 * unless debug printks are enabled, init_module and cleanup_module are silent,
 * with the exception of a memory mode announcement and a single success or failure message for each
 */


static int init_module(void)
{

	register struct ihcp_unit_t	*unit_p;		/* points to current unit structure     */
	int				instance;		/* current board #                      */
	int				unit_array_size;	/* size of array of unit structures     */
	u_long				temp;			/* generic temps                        */
	u_char				c_temp;
	u_long				bus_address;		/* temp base address                    */
	u_long				bus_base;		/* temp address base                    */
	u_long				bus_offset;		/* temp addres offset                   */
	u_long				wait_j;			/* wait jiffies                         */
	struct pci_dev			*temp_pci_dev_p;	/* temp pci dev struct pointer          */
	struct page			**maplist;		/* pointer to kiobuf page pointers	*/
	int				i;			/* handy loop index			*/

	/*
	 * tell the compiler and loader that we do not export any symbols
	 * all our static declarations will probably do that, but we will try to make sure...
	 */

	EXPORT_NO_SYMBOLS;

	temp = LINUX_VERSION_CODE;
	DPRINT((KERN_DEBUG
		"ihcp: init_module: compiled under version: %ld.%ld.%ld uts release: %s\n",
		(temp >> 16) & 0xffff, (temp >> 8) & 0xff, temp & 0xff, UTS_RELEASE));

	DPRINT((KERN_DEBUG "ihcp: init_module: entering init_module\n"));

	/*
	 * identify our memory mode/copy buffer status in msg log
	 */

	printk(KERN_INFO "ihcp: init_module: %s", IHCP_MEMORY_MODE_MSG);

	ihcp_module_initialized = 0;

	/*
	 * establish driver and board defaults - some of these will be used to establish initial configuration
	 * on a per-board basis
	 * 
	 * the config variables that end in _def may be later modified by ioctl calls on a per board basis
	 * 
	 * examine the variables that may be set during insmod - if not -1, and within permitted range, use the
	 * insmod supplied value - if -1, use the default
	 * 
	 * some of the range checks look redundant ( .. >= 0 ..), but I prefer to keep the insmod set detection
	 * separate from the range check, and not assume anything
	 */

	ihcp_byte_order_def = BYTE_ORDER_DEF;
	if (byte_order_def >= 0) {

		if ((byte_order_def == LOW_BYTE_FIRST)
		    || (byte_order_def == HIGH_BYTE_FIRST))
			ihcp_byte_order_def = byte_order_def;
		else
			printk(KERN_WARNING
			       "ihcp: init_module: byte_order_def out of range! using default\n");
	}

	DPRINT((KERN_DEBUG
		"ihcp: init_module: ihcp_byte_order_def = 0x%x\n",
		ihcp_byte_order_def));

	/*
	 * select one of the four versatec speeds as a default
	 */

	ihcp_vers_speed_def = VERS_SPEED_DEF;
	if (vers_speed_def >= 0) {

		if ((vers_speed_def >= 0) && (vers_speed_def <= 3))
			ihcp_vers_speed_def = vers_speed_def;
		else
			printk(KERN_WARNING
			       "ihcp: init_module: vers_speed_def out of range! using default\n");
	}

	DPRINT((KERN_DEBUG
		"ihcp: init_module: ihcp_vers_speed_def = 0x%x\n",
		ihcp_vers_speed_def));

	/*
	 * select one of the four centronics speeds as the default
	 */

	ihcp_cent_speed_def = CENT_SPEED_DEF;
	if (cent_speed_def >= 0) {

		if ((cent_speed_def >= 0) && (cent_speed_def <= 3))
			ihcp_cent_speed_def = cent_speed_def;
		else
			printk(KERN_WARNING
			       "ihcp: init_module: cent_speed_def out of range! using default\n");
	}

	DPRINT((KERN_DEBUG
		"ihcp: init_module: ihcp_cent_speed_def = 0x%x\n",
		ihcp_cent_speed_def));

	/*
	 * select power-up print/plot mode for versatec boards
	 * 0 = print, 1 = plot (probably should use #macros, not numeric values, but ...
	 */

	ihcp_mode_def = MODE_DEF;
	if (mode_def >= 0) {

		if ((mode_def == 0) || (mode_def == 1))
			ihcp_mode_def = mode_def;
		else
			printk(KERN_WARNING
			       "ihcp: init_module: mode_def out of range! using default\n");
	}

	DPRINT((KERN_DEBUG
		"ihcp: init_module: ihcp_mode_def = 0x%x\n",
		ihcp_mode_def));

	/*
	 * set default time to wait for dma to complete
	 */

	ihcp_dma_time_def = DMA_TIME_DEF;
	if (dma_time_def >= 0) {

		if ((dma_time_def >= DMA_TIME_MIN)
		    && (dma_time_def <= DMA_TIME_MAX)) ihcp_dma_time_def =
			    dma_time_def;
		else
			printk(KERN_WARNING
			       "ihcp: init_module: dma_time_def out of range! using default\n");
	}

	DPRINT((KERN_DEBUG
		"ihcp: init_module: ihcp_dma_time_def = 0x%x\n",
		ihcp_dma_time_def));

	/*
	 * set default time to wait for fifo empty and board and device ready
	 */

	ihcp_fifo_time_def = FIFO_TIME_DEF;
	if (fifo_time_def >= 0) {

		if ((fifo_time_def >= FIFO_TIME_MIN)
		    && (fifo_time_def <=
			FIFO_TIME_MAX)) ihcp_fifo_time_def = fifo_time_def;
		else
			printk(KERN_WARNING
			       "ihcp: init_module: fifo_time_def out of range! using default\n");
	}

	DPRINT((KERN_DEBUG
		"ihcp: init_module: ihcp_fifo_time_def = 0x%x\n",
		ihcp_fifo_time_def));

	/*
	 * set max board number - this will determine how much we kmalloc for the unit structures
	 * we will do a sanity check, and require this number to be a single digit
	 */

	ihcp_max_boards = MAX_BOARDS_DEF;
	if (max_boards >= 0) {

		if ((max_boards >= 0) && (max_boards <= 9))
			ihcp_max_boards = max_boards;
		else
			printk(KERN_WARNING
			       "ihcp: init_module: max_boards out of range! using default\n");
	}

	DPRINT((KERN_DEBUG
		"ihcp: init_module: ihcp_max_boards = %d\n",
		ihcp_max_boards));

	/*
	 * set max_phys_order and several values
	 * that derive from it
	 * 
	 * max buffer pages = 2**max phys order
	 * max bytes per dma xfer = max buffer pages * PAGE_SIZE
	 * max size of scatter/gather list = max buffer pages + 1
	 *
	 * these calculations must track w/kiobuf size as well
	 * 
	 * size of iopb memory allocated = max sg size * IOPB_SIZE
	 * 
	 * we will check for a request less than zero, but not limit the upper end - USER BEWARE!
	 */

	ihcp_max_phys_order = MAX_PHYS_ORDER_DEF;
	if (max_phys_order >= 0)
		ihcp_max_phys_order = max_phys_order;

	ihcp_max_buf_pages = (1 << ihcp_max_phys_order);
	ihcp_max_buf_bytes = ihcp_max_buf_pages * PAGE_SIZE;
	ihcp_max_sg_length = ihcp_max_buf_pages + 1;
	ihcp_iopb_mem_size = ihcp_max_sg_length * IOPB_SIZE;

	DPRINT((KERN_DEBUG
		"ihcp: init_module: max_phys_order = 0x%x, max_bytes = 0x%x, max_sg_length = 0x%x, iopb_mem_size = 0x%x\n",
		ihcp_max_phys_order, ihcp_max_buf_bytes, ihcp_max_sg_length,
		ihcp_iopb_mem_size));


	/*
	 * allocate space for unit structure array
	 * get enough for ihcp_max_boards unit structures
	 * zero the allocated memory, and save a pointer
	 * to the array
	 */

	unit_array_size = ihcp_max_boards * sizeof(struct ihcp_unit_t);
	ihcp_unit_base_p =
	    (struct ihcp_unit_t *) kmalloc(unit_array_size, GFP_KERNEL);

	if (ihcp_unit_base_p == NULL) {
		printk(KERN_ERR
		       "ihcp: init_module: can't allocate unit struct array (kmalloc error)!\n");
		return -ENOMEM;
	}

	memset(ihcp_unit_base_p, 0, unit_array_size);

	DPRINT((KERN_DEBUG
		"ihcp: init_module: allocated unit structure array, ihcp_unit_base_p = 0x%p, size = 0x%x\n",
		ihcp_unit_base_p, unit_array_size));


	/*
	 * make sure this is a pci machine
	 */

	if (!pci_present()) {
		printk(KERN_ERR
		       "ihcp: init_module: pci bus support not available! module not initialized!\n");
		kfree(ihcp_unit_base_p);
		return -ENODEV;
	}

	/*
	 * loop, looking for our boards - there better be at least one
	 */

	DPRINT((KERN_DEBUG
		"ihcp: init_module: starting search for boards\n"));

	ihcp_boards_found = 0;
	ihcp_boards_attached = 0;

	/*
	 * initialize pci dev struct pointer used to walk through device structures
	 */

	temp_pci_dev_p = NULL;

	for (instance = 0; instance < ihcp_max_boards; instance++) {

		/*
		 * make a pointer to the correct element of the unit structure array, and save the instance number
		 * the instance number  may be useful later - for debug and error reporting
		 */

		unit_p = ihcp_unit_base_p + instance;
		unit_p->instance = instance;

		/*
		 * look for board
		 */

		DPRINT((KERN_DEBUG
			"ihcp: init_module: probing for board at instance %d, unit_p = 0x%p\n",
			instance, unit_p));

		temp_pci_dev_p =
		    pci_find_device(IKON_VENDOR_ID, IHCP_DEVICE_ID,
				    temp_pci_dev_p);

		/*
		 * if successful, we have a board
		 * if not found, it means that we have already found all of our boards - IF ANY
		 * if anything else, we print an error message
		 * 
		 * there doesn't seem to be any way to detect an error, just a NULL result
		 */

		if (temp_pci_dev_p == NULL) {
			DPRINT((KERN_DEBUG
				"ihcp: init_module: instance %d: no (more) boards found, ending search\n",
				instance));
			break;
		}

		/*
		 * we have actually found a board - we have work to do!
		 * 
		 * we won't increment ihcp_boards_attached until we get through this without errors!!!
		 */

		/*
		 * save ptr to this board's pci dev structure
		 */

		unit_p->pci_dev_p = temp_pci_dev_p;


		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: board found, initializing unit structure\n",
			instance));

		ihcp_boards_found++;

		/*
		 * memset undoubtedly zero'ed our unit structure, but we will init some things in a more formal manner
		 * it seems like a good idea to refer to these struct elements explicitly - as a reminder, if nothing else
		 * 
		 * initialize mutex and conditional variable
		 */

		mutex_init(&unit_p->mutex);
		cond_init(&unit_p->cond);

		unit_p->sleeping = 0;
		unit_p->unit_attached = 0;
		unit_p->unit_open = 0;
		unit_p->unit_flags = 0;

		/*
		 * get contiguous pages of kernel memory to contain the iopbs (dma chaining list)
		 * GFP_KERNEL should get us contiguous memory, don't set dma flag so it won't be forced into low memory
		 */

		unit_p->iopb_base_p = kmalloc(ihcp_iopb_mem_size, GFP_KERNEL);

		if (unit_p->iopb_base_p == NULL) {
			printk(KERN_ERR
			       "ihcp: init_module: instance %d: kmalloc failure (iopb memory), board not attached!\n",
			       instance);
			continue;
		}

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: iopb memory allocated, virtual address = 0x%p\n",
			instance, unit_p->iopb_base_p));

		/*
		 * check for alignment - we could have asked for more and forced it, but I don't think the mem caches have any
		 * chunks w/smaller alignments than we need
		 */

		if ((u_long)unit_p->iopb_base_p & (IOPB_SIZE - 1)) {
			printk(KERN_ERR
			       "ihcp: init_module: instance %d: iopb memory not IOPB_SIZE aligned!\n",
			       instance);
			kfree(unit_p->iopb_base_p);
			continue;
		}

		/*
		 * inform the kernel what our dma addressing capabilities are
		 * we are a 32 bit device, and this should be the default, but
		 * we will try to be thorough
		 *
		 * this macro resolves to 0 if IHCP_USE_DMA_MASK is undefined
		 */

		if (IHCP_PCI_SET_DMA_MASK(temp_pci_dev_p, IHCP_DMA_MASK)) {
			printk(KERN_ERR
				"ihcp: init_module: instance %d: DMA addressing not compatible!\n",
				instance);
			kfree(unit_p->iopb_base_p);
			continue;
		}

		/*
		 * get the bus address of iopb memory using the dynamic DMA mapping stuff, (doesn't amount to much on
		 * a PC, but will help w/compatibility on SPARC, and other arch's w/iommu hardware
		 */

		unit_p->dma_handle =
		    pci_map_single(temp_pci_dev_p, unit_p->iopb_base_p,
				   ihcp_iopb_mem_size, PCI_DMA_TODEVICE);

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: iopb memory mapped, dma_handle = 0x%x\n",
			instance, (uint32_t)unit_p->dma_handle));

		/*
		 * get space for the scatterlist ... this isn't really necessary on a pc, since we should be
		 * able to convert the kiobuf page mappings direcly into bus adds, and build the iopb list,
		 * but to try for compatibility with other architectures, and to deal with CONFIG_HIGHMEM
		 * machines, we will go throug a scatterlist on the way to building an iopb list
		 *
		 * we now have to zero the allocated memory, since the kernel gods have added new
		 * .page and .offset to the scatterlist struct, and if .page AND .address are either
		 * both zero or both non-zero, they BUG()!
		 */

		unit_p->sg_list_p =
		    kmalloc(ihcp_max_sg_length *
			    sizeof(struct scatterlist), GFP_KERNEL);

		if (unit_p->sg_list_p == NULL) {
			printk(KERN_ERR
			       "ihcp: init_module: instance %d: can't allocate scatter list (kmalloc error)!\n",
			       instance);
			pci_unmap_single(temp_pci_dev_p,
					 unit_p->dma_handle,
					 ihcp_iopb_mem_size,
					 PCI_DMA_TODEVICE);
			kfree(unit_p->iopb_base_p);
			continue;
		}

		memset(unit_p->sg_list_p, 0, ihcp_max_sg_length * sizeof(struct scatterlist));

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: scatter list allocated, sg_list_p = 0x%p\n",
			instance, unit_p->sg_list_p));

		/*
		 * allocate a kiobuf for user buffer mapping
		 * it will arrive w/space in the map array for 64k of data
		 * we will expand it here to hold our max_buf_size  which requires
		 * an extra page - since the user buffer may well not be page aligned
		 *
		 * the underlying routine defined in the macro is selected by a compile
		 * option (this to tolerate RedHat 7.2 (updated) which uses -sz versions
		 * of the alloc and free routines
		 */

		if (IHCP_ALLOC_KIOVEC(1, &unit_p->kiobuf_p)) {
			printk(KERN_ERR
			       "ihcp: init_module: instance %d: can't allocate kiobuf!\n",
			       instance);
			pci_unmap_single(temp_pci_dev_p,
					 unit_p->dma_handle,
					 ihcp_iopb_mem_size,
					 PCI_DMA_TODEVICE);
			kfree(unit_p->sg_list_p);
			kfree(unit_p->iopb_base_p);
			continue;
		}

		if (expand_kiobuf(unit_p->kiobuf_p, ihcp_max_sg_length )) {
			printk(KERN_ERR
			       "ihcp: init_module: instance %d: can't expand kiobuf!\n",
			       instance);
			IHCP_FREE_KIOVEC(1, &unit_p->kiobuf_p);
			pci_unmap_single(temp_pci_dev_p,
					 unit_p->dma_handle,
					 ihcp_iopb_mem_size,
					 PCI_DMA_TODEVICE);
			kfree(unit_p->sg_list_p);
			kfree(unit_p->iopb_base_p);
			continue;
		}

		DPRINT((KERN_DEBUG "ihcp: init_module: instance %d: kiobuf allocated and expanded\n", instance));

		/*
		 * if a copy buffer is required, allocate it here and stuff the kiobuf w/its
		 * pages
		 *
		 * we only allocate just enough pages to satisfy the max buffer size, since we will always
		 * copy to the buffer starting at offset = 0, so we don't need an extra page to handle
		 * alignment (like the user kiobuf and sg list do)
		 */

		if (IHCP_COPY_BUF) {
			
			DPRINT((KERN_DEBUG "ihcp: init_module: instance %d: allocating copy buffer\n", instance));

			unit_p->dma_copy_buf_p = (char *)__get_free_pages(GFP_KERNEL,
					ihcp_max_phys_order);
			if(!unit_p->dma_copy_buf_p) {
				printk(KERN_ERR "ihcp: init_module: instance %d: can't allocate copy buffer!\n", instance);
				IHCP_FREE_KIOVEC(1, &unit_p->kiobuf_p);
				pci_unmap_single(temp_pci_dev_p,
						 unit_p->dma_handle,
						 ihcp_iopb_mem_size,
						 PCI_DMA_TODEVICE);
				kfree(unit_p->sg_list_p);
				kfree(unit_p->iopb_base_p);			
				continue;
			}

			DPRINT((KERN_DEBUG "ihcp: init_module: instance %d: copy buffer allocated, address = 0x%p\n",
					instance, unit_p->dma_copy_buf_p));

			/*
		 	 * get pointer to array of page struct pointers and fill them w/copy buff pages
			 * since we are in copy mode, these pointers will never change
			 */

			maplist = unit_p->kiobuf_p->maplist;
			for (i=0; i<ihcp_max_buf_pages; i++)
				maplist[i] = virt_to_page(unit_p->dma_copy_buf_p + ( i * PAGE_SIZE));

		} else {
		
			DPRINT((KERN_DEBUG "ihcp: init_module: instance %d: user DMA mode, no copy buffer allocated\n",
					instance));
			
			unit_p->dma_copy_buf_p = NULL;
		}

		/*
		 * grind through the config registers to get register base addresses and other values
		 * use the new pci interface, but keep the rest of the old code (lazy)
		 * 
		 * the base address registers contain the bus addresses of our two register sets - plx and ihcp
		 * unfortunately, the plx register set is smaller than a page, so bios may align it on an address
		 * that isn't a page boundary
		 * 
		 * ioremap doesn't tolerate that, so we have to figure out the appropriate (maybe lower) page boundary
		 * and the appropriate size to ask for
		 * we will save the mapped base and the actual virtual pointer to the register set
		 * 
		 * I hope linux allows multiple processes to map the same bus page, or we may clobber someone else
		 */


		/*
		 * make sure io and memory access is enabled
		 * make sure that the pci master enable bit is set
		 *
		 * pci_enable_device does some things on non-x86 arch's
		 * that must be done before accessing device resources registers
		 */

		pci_enable_device(temp_pci_dev_p);
		pci_set_master(temp_pci_dev_p);

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: master enable bits set\n",
			instance));

		bus_address = temp_pci_dev_p->resource[0].start;

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: pci_base_address_0 (plx regs) = 0x%lx\n",
			instance, bus_address));

		bus_address &= PCI_BASE_ADDRESS_MEM_MASK;
		bus_base = bus_address & ~(PAGE_SIZE - 1);
		bus_offset = bus_address & (PAGE_SIZE - 1);

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: plx bus add = 0x%lx, base = 0x%lx, offset = 0x%lx\n",
			instance, bus_address, bus_base, bus_offset));

		unit_p->plx_page_base_p = ioremap(bus_base, PLX_REG_SIZE + bus_offset);
		unit_p->plx_base_p = unit_p->plx_page_base_p + bus_offset;

		if (unit_p->plx_page_base_p == NULL) {
			printk(KERN_ERR
			       "ihcp: init_module: instance %d: can't ioremap plx registers!\n",
			       instance);
			if (unit_p->dma_copy_buf_p) free_pages((unsigned long)unit_p->dma_copy_buf_p,
					ihcp_max_phys_order);
			IHCP_FREE_KIOVEC(1, &unit_p->kiobuf_p);
			pci_unmap_single(temp_pci_dev_p,
					 unit_p->dma_handle,
					 ihcp_iopb_mem_size,
					 PCI_DMA_TODEVICE);
			kfree(unit_p->sg_list_p);
			kfree(unit_p->iopb_base_p);
			continue;
		}

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: plx mapped base = 0x%p, virt addr = 0x%p\n",
			instance, unit_p->plx_page_base_p, unit_p->plx_base_p));

		bus_address = temp_pci_dev_p->resource[2].start;

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: pci_base_address_2 (ihcp regs) = 0x%lx\n",
			instance, bus_address));

		bus_address &= PCI_BASE_ADDRESS_MEM_MASK;
		bus_base = bus_address & ~(PAGE_SIZE - 1);
		bus_offset = bus_address & (PAGE_SIZE - 1);

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: ihcp bus add = 0x%lx, base = 0x%lx, offset = 0x%lx\n",
			instance, bus_address, bus_base, bus_offset));

		unit_p->ihcp_page_base_p = ioremap(bus_base, IHCP_REG_SIZE + bus_offset);
		unit_p->ihcp_base_p = unit_p->ihcp_page_base_p + bus_offset;

		if (unit_p->ihcp_page_base_p == NULL) {
			printk(KERN_ERR
			       "ihcp: init_module: instance %d: can't ioremap ihcp registers!\n",
			       instance);
			iounmap(unit_p->plx_page_base_p);
			if (unit_p->dma_copy_buf_p) free_pages((unsigned long)unit_p->dma_copy_buf_p,
					ihcp_max_phys_order);
			IHCP_FREE_KIOVEC(1, &unit_p->kiobuf_p);
			pci_unmap_single(temp_pci_dev_p,
					 unit_p->dma_handle,
					 ihcp_iopb_mem_size,
					 PCI_DMA_TODEVICE);
			kfree(unit_p->sg_list_p);
			kfree(unit_p->iopb_base_p);
			continue;
		}

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: ihcp mapped base = 0x%p, virt addr = 0x%p\n",
			instance, unit_p->ihcp_page_base_p, unit_p->ihcp_base_p));

		/*
		 * get (possibly) mapped int level from pci_dev for use in kernel calls, and
		 * the hardware level from the board for use when restoring the level after
		 * an eeprom reload
		 */

		unit_p->pci_dev_int_level = temp_pci_dev_p->irq;

		if (pci_read_config_byte
		    (temp_pci_dev_p, PCI_INTERRUPT_LINE,
		     (u_char *) & c_temp)) {
			printk(KERN_ERR
			       "ihcp: init_module: instance %d: error reading pci interrupt line!\n",
			       instance);
			iounmap(unit_p->ihcp_page_base_p);
			iounmap(unit_p->plx_page_base_p);
			if (unit_p->dma_copy_buf_p) free_pages((unsigned long)unit_p->dma_copy_buf_p,
					ihcp_max_phys_order);
			IHCP_FREE_KIOVEC(1, &unit_p->kiobuf_p);
			pci_unmap_single(temp_pci_dev_p,
					 unit_p->dma_handle,
					 ihcp_iopb_mem_size,
					 PCI_DMA_TODEVICE);
			kfree(unit_p->sg_list_p);
			kfree(unit_p->iopb_base_p);
			continue;
		}

		unit_p->int_level = c_temp;

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: board int line = 0x%x, pci_dev int line = 0x%x\n",
			instance, unit_p->int_level,
			unit_p->pci_dev_int_level));

		/*
		 * for some reason we save device and vendor ids and rev level -
		 * we already know the id's or we wouldn't be here!
		 * 
		 * we could also pull some of this directly from temp_pci_dev_p
		 */

		if (pci_read_config_dword
		    (temp_pci_dev_p, PCI_VENDOR_ID, (int *) &temp)) {
			printk(KERN_ERR
			       "ihcp: init_module: instance %d: error reading pci vendor id! board not attached! \n",
			       instance);
			iounmap(unit_p->ihcp_page_base_p);
			iounmap(unit_p->plx_page_base_p);
			if (unit_p->dma_copy_buf_p) free_pages((unsigned long)unit_p->dma_copy_buf_p,
					ihcp_max_phys_order);
			IHCP_FREE_KIOVEC(1, &unit_p->kiobuf_p);
			pci_unmap_single(temp_pci_dev_p,
					 unit_p->dma_handle,
					 ihcp_iopb_mem_size,
					 PCI_DMA_TODEVICE);
			kfree(unit_p->sg_list_p);
			kfree(unit_p->iopb_base_p);
			continue;
		}

		unit_p->dev_and_vendor_id = temp;

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: device and vendor id = 0x%lx\n",
			instance, temp));


		if (pci_read_config_byte
		    (temp_pci_dev_p, PCI_REVISION_ID, (char *) &temp)) {
			printk(KERN_ERR
			       "ihcp: init_module: instance %d: error reading revision id! board not attached!\n",
			       instance);
			iounmap(unit_p->ihcp_page_base_p);
			iounmap(unit_p->plx_page_base_p);
			if (unit_p->dma_copy_buf_p) free_pages((unsigned long)unit_p->dma_copy_buf_p,
					ihcp_max_phys_order);	
			IHCP_FREE_KIOVEC(1, &unit_p->kiobuf_p);
			pci_unmap_single(temp_pci_dev_p,
					 unit_p->dma_handle,
					 ihcp_iopb_mem_size,
					 PCI_DMA_TODEVICE);
			kfree(unit_p->sg_list_p);
			kfree(unit_p->iopb_base_p);
			continue;
		}

		unit_p->revision_id = temp & 0xff;

		DPRINT((KERN_DEBUG
			"ihcp: init_modlue: instance %d: revision id = 0x%lx\n",
			instance, temp & 0xff));

		/*
		 * clear the plx and ikon logic by doing a soft reset
		 * this may be appropriate if module being loaded following an unload with no reboot
		 * 
		 * we don't do this at open, close, or unload time, since there might still be data in the fifo
		 * that hasn't gotten to the plotter yet (esp. at open and close time)
		 * 
		 * soft reset clears the local config registers, which turns off add space 0 accesses
		 * 
		 * so we will force an eeprom reload, and wait a while (1s) to avoid too many retry cycles
		 * while the eeprom is reloading - we will use schedule, not udelay, so we don't
		 * prevent other processing from doing useful things
		 * 
		 * THIS MUST NOT BE DONE FROM THE INTERRUPT SIDE!!!!!
		 * 
		 * do a read dump to flush any write buffer, then wait before touching the plx chip
		 * it returns retry status during config reload, and a write buffer or bridge chip
		 * with a short timeout could cause a panic
		 * 
		 * an earlier rev cleared the int line byte in the configuration registers during
		 * soft reset, but that is not supposed to be true anymore
		 * 
		 * IT SEEMS THAT IT IS STILL TRUE - SO WE WILL USE CONFIG REG ACCESS TO SAVE THE
		 * INT LINE BEFORE THE SOFT RESET, AND RESTORE IT AFTER
		 * 
		 * we probably don't have to clear the config reload bit after writing it,
		 * but we will to be sure
		 */


		PLX_PUTL(PLX_EEPROM_USER,
			 (PLX_GETL(PLX_EEPROM_USER) | PLX_SOFT_RESET));
		PLX_PUTL(PLX_EEPROM_USER,
			 (PLX_GETL(PLX_EEPROM_USER) & ~PLX_SOFT_RESET));

		read_dump_0 = PLX_GETL(PLX_INT_CSTAT);

		PLX_PUTL(PLX_EEPROM_USER,
			 (PLX_GETL(PLX_EEPROM_USER) |
			  CONFIGURATION_RELOAD));

		/*
		 * wait ...
		 */

		wait_j = jiffies + 1 * HZ;

		while (time_before(jiffies, wait_j))
			schedule();

		PLX_PUTL(PLX_EEPROM_USER,
			 (PLX_GETL(PLX_EEPROM_USER) &
			  ~CONFIGURATION_RELOAD));

		pci_write_config_byte(temp_pci_dev_p, PCI_INTERRUPT_LINE,
				      unit_p->int_level);

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: soft reset and configuration reload complete\n",
			instance));

		/*
		 * soft reset leaves pci master int enable on, so turn it off for safety
		 * don't bother with oring bits - just hammer them all off!
		 */


		PLX_PUTL(PLX_INT_CSTAT, 0);

		/*
		 * disable address space 0 pre-fetch
		 */

		PLX_PUTL(PLX_ADD_0_ROM_DESC, PLX_GETL(PLX_ADD_0_ROM_DESC) | ADD_0_PREFETCH_DISABLE);

		/*
		 * get a plx register and dprint it for a sanity check - make sure we can get to plx regs
		 */

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: sanity check: PLX EEPROM reg = 0x%x (s.b. xxxx767e)\n",
			instance, PLX_GETL(PLX_EEPROM_USER)));

		/*
		 * get the board strapping and save it
		 * zeros rest of flag register at same time
		 */

		unit_p->unit_flags =
		    INTERFACE_STRAP_MASK & IHCP_GETL(IHCP_DEVICE_STATUS);

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: sanity check: interface strapping = 0x%lx\n",
			instance, IHCP_BOARD_MASK & unit_p->unit_flags));


		/*
		 * set the per-board default configuration - derived from the global default variables
		 * 
		 * this rather elaborate multi-layer default scheme is left over from earlier drivers
		 * the use of "#defined" values allows for different underlying register layouts and bit definitins
		 * without changing the structure of the driver or the values that insmod can pass
		 */


		switch (ihcp_vers_speed_def) {

		case 0:
			unit_p->vers_speed_def = SPEED_0;
			break;

		case 1:
			unit_p->vers_speed_def = SPEED_1;
			break;

		case 2:
			unit_p->vers_speed_def = SPEED_2;
			break;

		case 3:
			unit_p->vers_speed_def = SPEED_3;
			break;

		}

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: versatec speed default set to 0x%lx\n",
			instance, unit_p->vers_speed_def));


		switch (ihcp_cent_speed_def) {

		case 0:
			unit_p->cent_speed_def = SPEED_0;
			break;

		case 1:
			unit_p->cent_speed_def = SPEED_1;
			break;

		case 2:
			unit_p->cent_speed_def = SPEED_2;
			break;

		case 3:
			unit_p->cent_speed_def = SPEED_3;
			break;

		}

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: centronics speed default set to 0x%lx\n",
			instance, unit_p->cent_speed_def));


		if (ihcp_mode_def == 1)
			unit_p->mode_def = PLOT_MODE;
		else
			unit_p->mode_def = NORM_PRINT_MODE;

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: mode default set to 0x%lx\n",
			instance, unit_p->mode_def));


		unit_p->dma_time_def = ihcp_dma_time_def;

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: dma time default set to 0x%lx\n",
			instance, unit_p->dma_time_def));


		unit_p->fifo_time_def = ihcp_fifo_time_def;

		DPRINT((KERN_DEBUG
			"ihcp: init_module: instance %d: fifo time default set to 0x%lx\n",
			instance, unit_p->fifo_time_def));

		/*
		 * set the data byte ordering default mode
		 * the native ordering for the board is low byte first
		 * 
		 * if high byte first mode is selected, the single byte data out and command out
		 * bytes must be sent to the fifos in the HIGH byte of the longword!
		 * 
		 * ordering is set here, and stored in the unit structure, since it will affect
		 * all later accesses to the 8 bit data and command registes
		 * 
		 * we only take action if high_byte_first is selected, since the preceeding reset
		 * should have set us to low_byte_first mode
		 * 
		 * setting high byte first mode requires turning on the byte_swizzle bit in the
		 * ihcp logic, and TURNING OFF the user output bit in the plx logic
		 */

		if (ihcp_byte_order_def == HIGH_BYTE_FIRST) {
			DPRINT((KERN_DEBUG
				"ihcp: init_module: instance %d: setting ordering mode to high byte first\n",
				instance));
			IHCP_PUTL(IHCP_MODE,
				  (IHCP_GETL(IHCP_MODE) | BYTE_SWIZZLE));
			PLX_PUTL(PLX_EEPROM_USER,
				 (PLX_GETL(PLX_EEPROM_USER) &
				  ~USER_OUTPUT));
		} else
			DPRINT(
			       (KERN_DEBUG
				"ihcp: init_module: instance %d: ordering mode left at low byte first\n",
				instance));

		unit_p->byte_order = ihcp_byte_order_def;

		/*
		 * write the default handshake speeds for versatec and centronics boards (as determined by strapping),
		 * respectively, and the default print/plot mode for versatec boards
		 * the print/plot mode is also saved in the unit structure, since it may be referred
		 * to later - and isn't visible in the board's registers
		 * 
		 * NOTE THAT THE SAME DEFAULTS ARE USED FOR ALL BOARDS OF THE SAME TYPE!!!!!!
		 */

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) != IHCPIO_CENT) {

			/*
			 * versatec - set speed and print/plot mode
			 */

			DPRINT((KERN_DEBUG
				"ihcp: init_module: instance %d: setting default versatec speed (on board) to 0x%lx\n",
				instance, unit_p->vers_speed_def));
			DPRINT(
			       (KERN_DEBUG
				"ihcp: init_module: instance %d: setting default versatec print/plot mode (on board) to 0x%lx\n",
				instance, unit_p->mode_def));

			IHCP_PUTL(IHCP_MODE,
				  ((IHCP_GETL(IHCP_MODE) & ~SPEED_MASK) |
				   unit_p->vers_speed_def));

			temp = SET_VERSATEC_MODE | unit_p->mode_def;
			if (unit_p->byte_order == HIGH_BYTE_FIRST)
				temp = temp << 24;
			IHCP_PUTL(IHCP_COMMAND_OUT, temp);

			unit_p->mode = unit_p->mode_def;

		} else {

			/*
			 * centronics - set speed only
			 */

			DPRINT((KERN_DEBUG
				"ihcp: init_module: instance %d: setting default centronics speed (on board) to 0x%lx\n",
				instance, unit_p->cent_speed_def));

			IHCP_PUTL(IHCP_MODE,
				  ((IHCP_GETL(IHCP_MODE) & ~SPEED_MASK) |
				   unit_p->cent_speed_def));
		}

		/*
		 * set saved data streaming mode - for old vme getregs call compatibility, since
		 * we can't read data streaming status from the logic on the other side of the
		 * fifo in the pci card
		 */

		unit_p->data_stream_mode = DATA_STREAMING_OFF;

		/*
		 * register the interrupt line for this board & link to our interrupt handler
		 * pass our unit pointer to the interrupt code
		 * 
		 * we will use our node name as a label for the interrupt - that way /proc/interrupts
		 * will haev a unique name per interrupt (per board) rather than all of them being
		 * named ihcp (we could use this when mknod is invoked)
		 * 
		 * we have to keep the name in the unit structure, since request_irq just saves the pointer,
		 * and doesn'g copy the string - so by the time /proc/interrupts gets read, the stack has
		 * changed, and the name would be garbage if not kept in "permanent" memory
		 * 
		 * (thank you Mr. Rubini!)
		 */

		(void) sprintf((char *) unit_p->minor_node_name, "ihcp%d", instance);

		if (request_irq
		    (unit_p->pci_dev_int_level, ihcp_intr, SA_SHIRQ,
		     (const char *) unit_p->minor_node_name,
		     (void *) unit_p)) {
			printk(KERN_ERR
			       "ihcp: init_module: instance %d: error requesting interrupt! board not attached! \n",
			       instance);
			iounmap(unit_p->ihcp_page_base_p);
			iounmap(unit_p->plx_page_base_p);
			if (unit_p->dma_copy_buf_p) free_pages((unsigned long)unit_p->dma_copy_buf_p,
					ihcp_max_phys_order);
			IHCP_FREE_KIOVEC(1, &unit_p->kiobuf_p);
			pci_unmap_single(temp_pci_dev_p,
					 unit_p->dma_handle,
					 ihcp_iopb_mem_size,
					 PCI_DMA_TODEVICE);
			kfree(unit_p->sg_list_p);
			kfree(unit_p->iopb_base_p);
			continue;
		}

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) == IHCPIO_CENT)
			printk(KERN_NOTICE
			       "ihcp: init_module: instance %d: board attached, irq = 0x%x, strapped for CENTRONICS\n",
			       instance, unit_p->pci_dev_int_level);
		else
			printk(KERN_NOTICE
			       "ihcp: init_module: instance %d: board attached, irq = 0x%x, strapped for VERSATEC\n",
			       instance, unit_p->pci_dev_int_level);


		/*
		 * set per-board attach complete flag for use in open
		 * and bump the count of successfully attached boards
		 */

		unit_p->unit_attached = 1;
		ihcp_boards_attached++;


	}			/* end of for (instance) loop   */

	/*
	 * if no boards have attached, give back the unit structure array and return an error
	 * if no boards are found, do the same, but print a different message
	 */

	if (!ihcp_boards_attached) {
		if (!ihcp_boards_found)
			printk(KERN_NOTICE
			       "ihcp: init_module: no boards found! module not initialized!\n");
		else
			printk(KERN_WARNING
			       "ihcp: init_module: %d board(s) found but no boards attached! module not initialized!\n",
			       ihcp_boards_found);

		kfree(ihcp_unit_base_p);
		return -ENODEV;
	}

	/*
	 * register the module, and get dynamic major number assigned
	 */

	ihcp_device_major = register_chrdev(0, "ihcp", &ihcp_fops);

	if (ihcp_device_major < 0) {
		printk(KERN_ERR
		       "ihcp: init_module: can't register character driver module! module not initialized!\n");

		/*
		 * call routine to walk through unit structure array and free all resources owned by attached boards
		 * and give back the unit array when done
		 */

		ihcp_free_units();
		kfree(ihcp_unit_base_p);
		return -ENODEV;
	}

	DPRINT((KERN_DEBUG
		"ihcp: init_module: driver registered, major # = %d\n",
		ihcp_device_major));

	/*
	 * flag module as initialized and return good status to insmod
	 */

	ihcp_module_initialized = 1;

	printk(KERN_NOTICE
	       "ihcp: init_module: %d board(s) found, %d board(s) attached, module initialized\n",
	       ihcp_boards_found, ihcp_boards_attached);

	return 0;

}				/* end of init_module   */



/*
 * ihcp_free_units walks through the allocated unit structure array and frees all the
 * resources of any board that was successfully attached, including iopb memory,
 * scatterlist, kiobuf, register mappings, and interrupt lines
 * 
 * it looks at all the slots, up to max_boards, not up to boards_attached, because it is
 * theoretically possible for pcibios errors to cause some boards to be found, but not
 * attach (VERY UNLIKELY)
 * 
 * THIS ROUTINE DOES NOT RETURN THE UNIT STRUCTURE ARRAY - THAT IS LEFT TO THE CALLING ROUTINE
 */


static void ihcp_free_units(void)
{

	register struct	ihcp_unit_t	*unit_p;	/* points at current unit structure     */
	int				instance;	/* current instance (minor number)      */


	/*
	 * look at each unit structure element of array
	 */

	DPRINT((KERN_DEBUG
		"ihcp_free_units: max boards = %d, boards found = %d, boards attached = %d\n",
		ihcp_max_boards, ihcp_boards_found, ihcp_boards_attached));

	for (instance = 0; instance < ihcp_max_boards; instance++) {

		unit_p = ihcp_unit_base_p + instance;

		if (unit_p->unit_attached) {
			DPRINT((KERN_DEBUG
				"ihcp_free_units: instance %d: freeing interrupt, mappings, iovec, copy buffer (if any), scatterlist, and iopb memory\n",
				instance));
			free_irq(unit_p->pci_dev_int_level,
				 (void *) unit_p);
			iounmap(unit_p->ihcp_page_base_p);
			iounmap(unit_p->plx_page_base_p);
			if (unit_p->dma_copy_buf_p) free_pages((unsigned long)unit_p->dma_copy_buf_p,
					ihcp_max_phys_order);
			IHCP_FREE_KIOVEC(1, &unit_p->kiobuf_p);
			pci_unmap_single(unit_p->pci_dev_p,
					 unit_p->dma_handle,
					 ihcp_iopb_mem_size,
					 PCI_DMA_TODEVICE);
			kfree(unit_p->sg_list_p);
			kfree(unit_p->iopb_base_p);
			unit_p->unit_attached = 0;
		} else
			DPRINT((KERN_DEBUG
				"ihcp_free_units: instance %d: this instance not attached - no action taken\n",
				instance));

	}			/* end of for (instance)        */

}				/* end of ihcp_free_units       */





/*
 * When our driver is unloaded, cleanup_module cleans up and frees the resources
 * we allocated in init_module
 * 
 * unregister the driver
 * 
 * return resources on an per instance basis (interrupts, kiovec, iobp stuff, etc)
 * 
 * free the unit structure array
 */


static void cleanup_module(void)
{
	register struct ihcp_unit_t	*unit_p;		/* will point to a particular unit structure    */
	int				instance;		/* board's minor number                         */
	int				boards_detached;	/* count of boards detached                     */


	DPRINT((KERN_DEBUG
		"ihcp: cleanup_module: entering cleanup_module\n"));

	/*
	 * make sure module was previously initialized
	 * there doesn't seem to be any way to return an error code
	 */

	if (!ihcp_module_initialized) {
		printk(KERN_ERR
		       "ihcp: cleanup_module: cleanup called when module not previously initialized!");
		return;
	}

	/*
	 * unregister the device (driver)
	 */

	if (unregister_chrdev(ihcp_device_major, "ihcp")) {
		printk(KERN_ERR
		       "ihcp: cleanup_module: unregister_chrdev error! exiting without cleaning up module!\n");
		return;
	}

	DPRINT((KERN_DEBUG
		"ihcp: cleanup_module: device (driver) successfully unregistered\n"));

	/*
	 * walk through the unit structure, and force any attached boards to be quiet (should already be)
	 * 
	 * we could have put this code in free_units, but the tasks are slightly different
	 */

	boards_detached = 0;

	for (instance = 0; instance < ihcp_max_boards; instance++) {

		unit_p = ihcp_unit_base_p + instance;

		if (unit_p->unit_attached) {

			DPRINT((KERN_DEBUG
				"ihcp: cleanup_module: instance %d: disabling dma and interrupts\n",
				instance));
			/*
			 * fist disable all plx interrupt enable bits
			 * then the ihcp masks
			 */

			PLX_PUTL(PLX_INT_CSTAT, 0);
			IHCP_PUTL(IHCP_INTERRUPT_MASK, 0);

			/*
			 * now make sure dma channel(s) are off
			 * for now we will access dma registers via the ihcp address window
			 * later chip revisions may require testing for rev level, and accessing
			 * via the plx map
			 * 
			 * we will just force all zeros into both channel's registers
			 */

			PLX_PUTL(PLX_DMA_CMD_STAT_BOTH, 0);

			boards_detached++;

		} else
			DPRINT((KERN_DEBUG
				"ihcp: cleanup_modlue: instance %d: this instance not attached - no action taken\n",
				instance));

	}			/* end of for (instance)        */

	/*
	 * call ihcp_free_units to release resources
	 * and turn off attached flag(s)
	 * 
	 * return unit structure memory
	 */

	ihcp_free_units();

	DPRINT((KERN_DEBUG "ihcp: cleanup_module: freeing unit structure\n"));

	kfree(ihcp_unit_base_p);

	/*
	 * clear module initialized flag
	 */

	ihcp_module_initialized = 0;

	printk(KERN_NOTICE "ihcp: cleanup_module: %d board(s) detached\n",
	       boards_detached);

	/*
	 * this is as good a place as any to do phony references to the read dump varaibles
	 * to try to absolutely guarantee that the optimizer won't remove our cache flush reads
	 * 
	 * and to try to shut lint up
	 */

	read_dump_0 += read_dump_1;

	return;

}



/*
 * do various error checking, then make the board available for write and ioctl calls
 * 
 * bump the usage count so driver can't be unloaded until close called
 */

static int ihcp_open(struct inode *inode_p, struct file *file_p)
{
	int				instance;
	register struct ihcp_unit_t	*unit_p;


	instance = MINOR(inode_p->i_rdev);

	DPRINT((KERN_DEBUG
		"ihcp_open: instance %d: entering ihcp_open\n", instance));

	/*
	 * make sure minor number is within range
	 */

	if ((instance < 0) || (instance >= ihcp_max_boards)) {
		printk(KERN_ERR
		       "ihcp_open: instance %d: instance out of range!\n",
		       instance);
		return -ENODEV;
	}

	/*
	 * get unit structure pointer and
	 * make sure driver is (completely) attached
	 */

	unit_p = ihcp_unit_base_p + instance;

	DPRINT((KERN_DEBUG
		"ihcp_open: instance %d: unit pointer = 0x%p\n", instance,
		unit_p));

	if (unit_p->unit_attached == 0) {
		printk(KERN_WARNING
		       "ihcp open: instance %d: open attempted when board not attached!\n",
		       instance);
		return -ENODEV;
	}

	/*
	 * Only allow a single open
	 */

	if (unit_p->unit_open != 0) {
		DPRINT((KERN_DEBUG
			"ihcp_open: instance %d: open called when already open!\n",
			instance));
		return -EBUSY;
	}

	/*
	 * mark device as open
	 * and increment usage count (if !=0, can't unload driver)
	 */

	unit_p->unit_open = 1;
	MOD_INC_USE_COUNT;

	/*
	 * clear unit flags - but not board type
	 */

	unit_p->unit_flags &= IHCP_CLEAR_FLAGS;

	/*
	 * set default timeouts here
	 * timeout parameters are stored as ticks
	 */

	unit_p->dma_time = (u_long) (HZ * unit_p->dma_time_def);
	unit_p->fifo_time = (u_long) (HZ * unit_p->fifo_time_def);

	DPRINT((KERN_DEBUG
		"ihcp_open: instance %d: open complete\n", instance));

	return 0;

}



/*
 * make the board unavailable for write and ioctl calls
 * 
 * decrement use count, so driver can be unloaded if desired
 */


static int ihcp_close(struct inode *inode_p, struct file *file_p)
{

	int				instance;
	register struct ihcp_unit_t	*unit_p;


	instance = MINOR(inode_p->i_rdev);

	DPRINT((KERN_DEBUG
		"ihcp_close: instance %d: entering ihcp_close\n",
		instance));

	/*
	 * make sure minor number is within range
	 */

	if ((instance < 0) || (instance >= ihcp_max_boards)) {
		printk(KERN_ERR
		       "ihcp_close: instance %d: instance out of range!\n",
		       instance);
		return -ENODEV;
	}

	/*
	 * get unit structure pointer
	 * make sure we are trying to close board that has already been opened
	 */

	unit_p = ihcp_unit_base_p + instance;

	DPRINT((KERN_DEBUG
		"ihcp_close: instance %d: unit pointer = 0x%p\n", instance,
		unit_p));

	if (unit_p->unit_open == 0) {
		printk(KERN_WARNING
		       "ihcp close: instance %d: close attempted when board not open!\n",
		       instance);
		return -ENODEV;
	}

	/*
	 * let the next user open the device
	 * and decrement use count
	 */

	unit_p->unit_open = 0;
	MOD_DEC_USE_COUNT;


	/*
	 * we don't reset the board and device here - that might kill a job that
	 * was in the device but not yet complete - reset in detach
	 */


	DPRINT((KERN_DEBUG
		"ihcp_close: instance %d: close complete\n", instance));

	return 0;

}



/*
 * do dma data transfer to device - and leading and trailing bytes in p-i/o mode, if necessary
 * in solaris and hpux drivers, the leading and trailing bytes were handled in strategy - we
 * do them in write in this driver (for some reason)
 * 
 * this is the version that we have all been waiting for -- it uses a kiobuf to pin and map a
 * user buffer so we can to direct dma UNLESS this is a CONFIG_HIGHMEM machine, and IHCP_HIGHMEM
 * is not defined, in which case we copy to a local kernel buffer, and the kiobuf is pre-initialized
 * to contain _those_ pages (pinning and mapping is not done - or necessary)
 * 
 * the size of the area mapped is limited by the ihcp_max_phys_order parameter - if the user asks for
 * more than this, it will be broken into chunks (xparent to the user), and done w/multiple
 * strategy calls
 * 
 * we use the dynamic DMA mapping functions here (although they aren't typically necessary on a PC) to
 * try for some kind of compatibility across architectures and allow for CONFIG_HIGHMEM machines
 * 
 * we wait for dma completion in strategy, rather than starting dma and returning, and
 * waiting for completion at the start of the next strategy call, to keep some kind of similarity
 * with our Solaris and HPUX drivers.
 * 
 * NOTE THAT THERE ARE LOTS OF ERROR RETURNS IN THIS CODE -- IF WE EVER NEED A MUTEX HERE, OR ITS EQUIVALENT,
 * THESE WILL ALL NEED TO SET AN ERROR CODE AND GOTO ... TO AVOID LEAKING OUT THE SIDE OF THE MUTEX
 */

static ssize_t ihcp_write(struct file *file_p, const char *user_buf_p,
			  size_t user_count, loff_t * off_p)
{

	struct inode			*inode_p;		/* will point to inode                  */
	register struct ihcp_unit_t	*unit_p;		/* points at our unit structure         */
	int				retval;			/* value returned from write            */
	int				instance;		/* our instance (minor) number          */
	int				dma_count;		/* even quad byte dma count             */
	u_char				temp_char;		/* char copy buffer                     */
	int				strategy_retval;	/* value returned from strategy         */

	/*
	 * we always say we xferred all the bytes, unless error
	 * the plx chip can't read back the dma count anyway
	 * 
	 * even if it could, we don't know how many bytes are still in fifos at timeout
	 */

	retval = user_count;

	/*
	 * extract inode pointer from file struct - we used to get this passed directly with write call - pre 2.2
	 */

	inode_p = file_p->f_dentry->d_inode;

	/*
	 * get instance and make unit struct pointer
	 */

	instance = MINOR(inode_p->i_rdev);

	DPRINT((KERN_DEBUG
		"ihcp_write: instance %d: entering write routine, user_buf_p = 0x%p, count = 0x%x\n",
		instance, user_buf_p, (unsigned int)user_count));

	if ((instance < 0) || (instance >= ihcp_max_boards)) {
		printk(KERN_ERR
		       "ihcp_write: instance %d: instance out of range!\n",
		       instance);
		return -ENODEV;
	}

	unit_p = ihcp_unit_base_p + instance;

	/*
	 * check - probably needlessly - to make sure we have been opened
	 */

	if (!unit_p->unit_open) {
		printk(KERN_ERR
		       "ihcp_write: instance %d: write called when unit not open!\n",
		       instance);
		return -EIO;
	}

	/*
	 * check user buffer pointer for leading bytes not quad byte aligned, and output them, up to user_count (max)
	 * bump user buf pointer and dec user count as bytes are sent
	 */

	while (((u_long) user_buf_p & 0x3) && (user_count)) {

		if (copy_from_user
		    (&temp_char, user_buf_p, sizeof(temp_char))) {
			printk(KERN_ERR
			       "ihcp_write: instance %d: copy_from_user error!\n",
			       instance);
			return -EINVAL;
		}

		DPRINT((KERN_DEBUG
			"ihcp_write: instance %d: sending initial byte = 0x%x, buf_p = 0x%p, count = 0x%x\n",
			instance, temp_char, user_buf_p, (unsigned int)user_count));

		/*
		 * test for fifo full while cranking out bytes - should never happen
		 */

		if ((IHCP_GETL(IHCP_INTERFACE_STATUS) & FIFO_NOT_FULL) ==
		    0) {
			printk(KERN_WARNING
			       "ihcp_write: instance %d: fifo full while sending leading bytes!\n",
			       instance);
			return -EIO;
		}

		if (unit_p->byte_order == HIGH_BYTE_FIRST)
			temp_char = temp_char << 24;
		IHCP_PUTL(IHCP_8_BIT_DATA_OUT, temp_char);

		user_buf_p++;
		user_count--;

	}			/* end of while */

	/*
	 * we should have an aligned user buffer pointer at this point, and 0 or more bytes left
	 * 
	 * clear the timeout and sig rx'd flags -- also clears "waiting for" flags
	 * strategy will set the "waiting for dma" flag.
	 */

	unit_p->unit_flags &= IHCP_CLEAR_FLAGS;

	/*
	 * loop on  dma out while user count >= 4
	 * use do..while loop so we go through at least once
	 * this will cause strategy to wait for fifo <1/2 full, even if no
	 * dma output to do.....protects against many non-dma buffers filling
	 * fifo
	 * 
	 * strategy expects the virtual address of the next chunk of user buffer, and a dma count
	 * that is 0 mod 4 (and may be 0)
	 */

	do {

		/*
		 * use the min of remaining user count and max dma buffer size (masked to 0 mod 4)
		 * as the dma count
		 */

		dma_count =
		    (user_count <
		     ihcp_max_buf_bytes) ? user_count : ihcp_max_buf_bytes;

		dma_count &= 0xfffffffc;

		DPRINT((KERN_DEBUG
			"ihcp_write: instance %d: calling strategy, user_buf_p = 0x%p, dma_count = 0x%x\n",
			instance, user_buf_p, dma_count));

		strategy_retval =
		    ihcp_strategy(unit_p, user_buf_p, dma_count);

		DPRINT((KERN_DEBUG
			"ihcp_write: instance %d: ihcp_strategy returns 0x%x\n",
			instance, strategy_retval));

		if (strategy_retval) {
			printk(KERN_WARNING
			       "ihcp_write: instance %d: ihcp_strategy returns error = %d!\n",
			       instance, strategy_retval);
			return strategy_retval;
		}

		/*
		 * bump user buffer pointer, and decrement user count
		 */

		user_buf_p += dma_count;
		user_count -= dma_count;

		DPRINT((KERN_DEBUG
			"ihcp_write: instance %d: at bottom of do..while, bytes remaining = 0x%x\n",
			instance, (unsigned int)user_count));

	} while (user_count >= 4);


	/*
	 * figure out if there are any trailing bytes, update user buf pointer and count
	 * and send the bytes via p-i/o
	 */


	while (user_count) {

		if (copy_from_user
		    (&temp_char, user_buf_p, sizeof(temp_char))) {
			printk(KERN_ERR
			       "ihcp_write: instance %d: copy_from_user error!\n",
			       instance);
			return -EINVAL;
		}

		DPRINT((KERN_DEBUG
			"ihcp_write: instance %d: sending trailing byte = 0x%x, buf_p = 0x%p, count = 0x%x\n",
			instance, temp_char, user_buf_p, (unsigned int)user_count));

		/*
		 * test for fifo full while cranking out bytes - should never happen
		 */

		if ((IHCP_GETL(IHCP_INTERFACE_STATUS) & FIFO_NOT_FULL) ==
		    0) {
			printk(KERN_WARNING
			       "ihcp_write: instance %d: fifo full while sending trailing bytes!\n",
			       instance);
			return -EIO;
		}

		if (unit_p->byte_order == HIGH_BYTE_FIRST)
			temp_char = temp_char << 24;
		IHCP_PUTL(IHCP_8_BIT_DATA_OUT, temp_char);

		user_buf_p++;
		user_count--;

	}			/* end of while */


	DPRINT((KERN_DEBUG
		"ihcp_write: instance %d: leaving write routine\n",
		instance));

	/*
	 * retval should be the original byte count if we get to this point
	 */

	return (retval);
}




/*
 * provide mechanism for controlling device and driver
 * 
 * THIS CODE USES A FEW ERROR RETURNS - IF WE SOMEDAY NEED A MUTEX OR EQUIVALENT, GO BACK TO
 * SOLARIS STYLE .. SET RETVAL ERROR CODES AND JUMP TO IOCTLEXIT, SO WE DON'T
 * RISK LEAKING OUT THE SIDE OF A MUTEX!
 * 
 * 
 */


static int ihcp_ioctl(struct inode *inode_p, struct file *file_p,
		      u_int cmd, u_long arg)
{

	register struct ihcp_unit_t *unit_p;		/* points at our unit struct    */

	u_long	bits;					/* will contain incoming or outgoing argument value     */
	u_char	ct;					/* temp char                                            */
	u_short	st;					/* temp short                                           */
	u_long	lt;					/* temp long                                            */
	int	count;					/* size of argument                                     */
	int	i, t;					/* temp working variables                               */
	int	instance;				/* our instance                                         */
	int	command;				/* ioctl command                                        */
	int	wait_return;				/* saved return value from conditional wait             */

	u_long	return_array[IHCP_RETURN_ARRAY_SIZE];	/* will contain copy of registers to return to caller - see ihcp_io.h   */
	u_char	data_array[256];			/* used to contain p-i/o data out values after copyin   */
	struct	lpregs lp_regs;				/* returns old-style versatec register status           */


	/*
	 * get instance and make unit struct pointer
	 */

	instance = MINOR(inode_p->i_rdev);

	DPRINT((KERN_DEBUG
		"ihcp_ioctl: instance %d: entering ioctl routine, command = 0x%x, arg = 0x%lx\n",
		instance, cmd, arg));

	/*
	 * make sure minor number is within valid range
	 */

	if ((instance < 0) || (instance >= ihcp_max_boards)) {
		printk(KERN_ERR
		       "ihcp_ioctl: instance %d: instance out of range!\n",
		       instance);
		return -ENODEV;
	}

	unit_p = ihcp_unit_base_p + instance;

	/*
	 * check - probably needlessly - to make sure we have been opened
	 */

	if (!unit_p->unit_open) {
		printk(KERN_ERR
		       "ihcp_ioctl: instance %d: ioctl called when unit not open!\n",
		       instance);
		return -EIO;
	}

	/*
	 * get command and argument size
	 */

	command = cmd & IHCPIO_CMD_MASK;
	count = (cmd & IHCPIO_COUNT_MASK) >> _IOC_SIZESHIFT;

	DPRINT((KERN_DEBUG
		"ihcp_ioctl: instance %d: masked command = 0x%x, argument size = 0x%x\n",
		instance, command, count));

	/*
	 * get first word of argument (even if we are eventually doing a
	 * copyout operation) - this saves some typeing
	 */

	if (copy_from_user(&bits, (caddr_t) arg, sizeof(bits))) {
		printk(KERN_ERR
		       "ihcp_ioctl: instance %d: copy_from_user error!\n",
		       instance);
		return -EINVAL;
	}

	DPRINT((KERN_DEBUG
		"ihcp_ioctl: instance %d: arg = 0x%lx, *arg = 0x%lx\n",
		instance, arg, bits));

	/*
	 * defice mask used to strip the count and upper control bits from cases
	 */

#define	MASK IHCPIO_CMD_MASK


	switch (command) {

	case MASK & IHCPIO_DEV_RESET:	/* send ~ 1us reset pulse to attached device */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at DEV_RESET\n",
			instance));

		/*
		 * set and reset device clear bit, with a delay so the pulse is long enough
		 * for even slow devices
		 * 
		 * udelay will make the cpu delay, but doesn't guarantee that the writes won't
		 * stack in a write buffer
		 * 
		 * so we will do a read before the wait to force the buffer to flush
		 */

		IHCP_PUTL(IHCP_INTERFACE_CONTROL,
			  (IHCP_GETL(IHCP_INTERFACE_CONTROL) |
			   RESET_DEVICE));
		read_dump_0 = IHCP_GETL(IHCP_MODE);
		udelay(1);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL,
			  (IHCP_GETL(IHCP_INTERFACE_CONTROL) &
			   ~RESET_DEVICE));

		break;

	case MASK & IHCPIO_SET_CONFIG:	/* set speed and if to use busy */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at SET_CONFIG\n",
			instance));

		/*
		 * mask to pertinent bits only
		 * the calling program will use the old sbus speed definitions
		 * we will have to convert them to the pci card bits here
		 */

		bits &=
		    (IHCPIO_V_BURST | IHCPIO_BUSY_HANDSHAKE | IHCPIO_4_EDGE
		     | IHCPIO_IGNORE_BUSY | IHCPIO_SPEED3);

		/*
		 * if request is a centronics function and board is strapped for versatec - error!
		 */

		if (
		    ((bits
		      & (IHCPIO_BUSY_HANDSHAKE | IHCPIO_4_EDGE |
			 IHCPIO_IGNORE_BUSY)) != 0)
		    && ((unit_p->unit_flags & IHCP_BOARD_MASK) !=
			IHCPIO_CENT)) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at SET_CONFIG - attempt to set centronics mode on board strapped for versatec!\n",
			       instance);
			return -ENOTTY;
		}

		/*
		 * if request is a versatec only function and board is strapped for centronics - error!
		 */

		if (((bits & IHCPIO_V_BURST) != 0) &&
		    ((unit_p->unit_flags & IHCP_BOARD_MASK) ==
		     IHCPIO_CENT)) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at SET_CONFIG - attempt to set versatec mode on baord straped for centronics!\n",
			       instance);
			return -ENOTTY;
		}

		/*
		 * shuffle bits to match new board
		 */

		t = SPEED_MASK & (3 - ((bits & IHCPIO_SPEED3) >> 2));
		if (bits & IHCPIO_BUSY_HANDSHAKE)
			t |= BUSY_NOT_ACK;
		if (bits & IHCPIO_4_EDGE)
			t |= FOUR_EDGE_HANDSHAKE;
		if (bits & IHCPIO_IGNORE_BUSY)
			t |= IGNORE_BUSY;
		if (bits & IHCPIO_V_BURST)
			t |= V_BURST;

		/*
		 * plug into mode register - being careful not to disturb other bits
		 */

		IHCP_PUTL(IHCP_MODE,
			  (IHCP_GETL(IHCP_MODE) &
			   (BYTE_SWIZZLE | REVERSE_DATA_ENB)) | t);

		break;

	case MASK & IHCPIO_SET_DMATIME:	/* set dma timeout parameter    */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at SET_DMATIME\n",
			instance));

		if (bits < DMA_TIME_MIN)
			bits = DMA_TIME_MIN;	/* not less than min */
		if (bits > DMA_TIME_MAX)
			bits = DMA_TIME_MAX;	/* or more than max */

		unit_p->dma_time = (u_long) (bits * HZ);	/* set ticks */

		break;

	case MASK & IHCPIO_SET_FIFOTIME:	/* set timeout for fifo <1/2 full */
						/* or fifo empty & dev rdy wait */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at SET_FIFOTIME\n",
			instance));

		if (bits < FIFO_TIME_MIN)
			bits = FIFO_TIME_MIN;	/* not less than */
		if (bits > FIFO_TIME_MAX)
			bits = FIFO_TIME_MAX;	/* or more than.. */

		unit_p->fifo_time = (u_long) (bits * HZ);	/* ticks */

		break;

	case MASK & IHCPIO_GET_REGS:	/* get all the board's registers */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at GET_REGS\n",
			instance));

		return_array[0] = unit_p->dev_and_vendor_id;		/* device and vendor id */
		return_array[1] = unit_p->revision_id;			/* revision id          */
		return_array[2] = PLX_GETL(PLX_INT_CSTAT);		/* interrupt control-status     */
		return_array[3] = PLX_GETL(PLX_EEPROM_USER);		/* eeprom control and user bits */
		return_array[4] = PLX_GETL(PLX_DMA_MODE_0);		/* dma mode reg 0       */
		return_array[5] = PLX_GETL(PLX_DMA_PCI_ADD_0);	/* dma pci add 0        */
		return_array[6] = PLX_GETL(PLX_DMA_LOC_ADD_0);	/* dma local add 0      */
		return_array[7] = PLX_GETL(PLX_DMA_COUNT_0);		/* dma transfer count 0 */
		return_array[8] = PLX_GETL(PLX_DMA_DESC_PTR_0);	/* dma descriptor pointer 0     */
		return_array[9] = PLX_GETL(PLX_DMA_CMD_STAT_BOTH);		/* dma command/status - both channels   */
		return_array[10] = IHCP_GETL(IHCP_INTERRUPT_MASK);	/* ihcp interrupt mask reg      */
		return_array[11] = IHCP_GETL(IHCP_MODE);		/* ihcp mode reg        */
		return_array[12] = IHCP_GETL(IHCP_DEVICE_CONTROL);	/* device control reg   */
		return_array[13] = IHCP_GETL(IHCP_INTERFACE_CONTROL);	/* interface control    */
		return_array[14] = IHCP_GETL(IHCP_INTERFACE_STATUS);	/* interface status     */
		return_array[15] = IHCP_GETL(IHCP_DEVICE_STATUS);	/* device status        */
		return_array[16] = IHCP_GETL(IHCP_REVERSE_DATA);	/* reverse data in reg  */
		return_array[17] = IHCP_GETL(IHCP_AUTO_LTR_LOW);	/* auto ltr count low byte      */
		return_array[18] = IHCP_GETL(IHCP_AUTO_LTR_HIGH);	/* auto ltr count high byte     */

		if (copy_to_user((caddr_t) arg, return_array, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_to_user error!\n",
			       instance);
			return -EINVAL;
		}

		break;

	case MASK & IHCPIO_GET_STATUS:	/* get formatted version of status */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at GET_STATUS\n",
			instance));

		/*
		 * get device status bits
		 */

		ct =
		    IHCP_GETL(IHCP_DEVICE_STATUS) & (FAULT | ONLINE |
						     PAPER_OUT |
						     CENTRONICS_BUSY |
						     VERSATEC_READY);

		/*
		 * shuffle status bits to match sbus format
		 */

		bits = 0;
		if (ct & FAULT)
			bits |= IHCPIO_DEV_FAULT;
		if (ct & ONLINE)
			bits |= IHCPIO_DEV_SEL;
		if (ct & PAPER_OUT)
			bits |= IHCPIO_DEV_POUT;
		if (ct & CENTRONICS_BUSY)
			bits |= IHCPIO_DEV_BUSY;
		if (ct & VERSATEC_READY)
			bits |= IHCPIO_DEV_RDY;

		if (copy_to_user((caddr_t) arg, &bits, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_to_user error!\n",
			       instance);
			return -EINVAL;
		}

		break;

	case MASK & IHCPIO_GET_BOARD:	/* get board type       */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at GET_BOARD\n",
			instance));

		bits =
		    IHCP_GETL(IHCP_DEVICE_STATUS) & INTERFACE_STRAP_MASK;

		if (copy_to_user((caddr_t) arg, &bits, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_to_user error!\n",
			       instance);
			return -EINVAL;
		}

		break;

	case MASK & IHCPIO_SET_VMODE:	/* set versatec mode - both leave rdy ff on on pci card */
	case MASK & IHCPIO_SET_VMODEX:

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at SET_VMODE\n",
			instance));

		/*
		 * make sure the board is strapped for versatec
		 */

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) == IHCPIO_CENT) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at SET_VMODE - attempt to do versatec mode set on baord strapped for centronics!\n",
			       instance);
			return -ENOTTY;
		}

		/*
		 * this command talks to the fifo - make sure there is room
		 * we could wait for room with an interrupt, but the way the board is to
		 * be used, there should always be room
		 */

		if ((IHCP_GETL(IHCP_INTERFACE_STATUS) & FIFO_NOT_FULL) ==
		    0) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at SET_VMODE - no room in fifo for versatec mode set!\n",
			       instance);
			return -EIO;
		}

		/*
		 * bottom two bits contain print/plot/spp control - same for sbus and pci
		 * command byte will be in low byte unless byte swizzling selected
		 */

		lt = SET_VERSATEC_MODE | (bits & 0x03);
		if (unit_p->byte_order == HIGH_BYTE_FIRST)
			lt = lt << 24;
		IHCP_PUTL(IHCP_COMMAND_OUT, lt);

		/*
		 * save mode in unit structure
		 */

		unit_p->mode = bits & 0x03;

		break;

	case MASK & IHCPIO_V_CMD:	/* issue versatec pulse command */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at V_CMD\n", instance));

		/*
		 * make sure the board is strapped for versatec - can't do this w/centr
		 */

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) == IHCPIO_CENT) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at V_CMD - attempt to issue versatec command to board strapped for centronics!\n",
			       instance);
			return -ENOTTY;
		}

		/*
		 * this command talks to the fifo - make sure there is room !!
		 */

		if ((IHCP_GETL(IHCP_INTERFACE_STATUS) & FIFO_NOT_FULL) ==
		    0) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at V_CMD - no room in fifo for versatec command!\n",
			       instance);
			return -EIO;
		}

		/*
		 * bottom two bits contain command - we have to shuffle them
		 * for compatibility with sbus driver
		 */

		if (bits == IHCPIO_VLTR)
			lt = VERSATEC_LINE_TERM;
		if (bits == IHCPIO_VEOT)
			lt = VERSATEC_EOT;
		if (bits == IHCPIO_VFED)
			lt = VERSATEC_FORM_FEED;
		if (bits == IHCPIO_VCLR)
			lt = VERSATEC_CLEAR;

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at V_CMD, command sent = 0x%lx\n",
			instance, lt));

		if (unit_p->byte_order == HIGH_BYTE_FIRST)
			lt = lt << 24;
		IHCP_PUTL(IHCP_COMMAND_OUT, lt);

		break;


	case MASK & IHCPIO_DATA_OUT:	/* output chars from arg - # of chars in bottom of cmd  */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at DATA_OUT, byte count = 0x%x\n",
			instance, count));

		/*
		 * make sure # of characterss is <=255
		 * that is the max # of characters in arg array - for historical reasons
		 * 
		 * we will check for room in fifo before each byte but will not wait on
		 * an interrupt if the fifo is full - we will flag an error and return
		 * 
		 * each byte will be sent by writing to the 8 bit data out register
		 * this is much less effecient than writing quad bytes to the 32 bit
		 * data out register, or to the burst out range
		 * 
		 * it is not the intent of this driver to use this ioctl for large data
		 * blocks, or frequent data transfers - those funtions should be served via
		 * the write call
		 * 
		 * this ioctl should be used for the output of a few bytes - perhaps as a
		 * header to a write call, and should be used between write calls, or waits for
		 * fifo less than half full
		 * 
		 * if this ioctl is used repeatedly, without interspersed dma transfers, which
		 * leave the fifo approximately half full, or without wait for half full ioctls
		 * the fifo may eventually overflow
		 * 
		 * it would be possible to incorporate a wait for less than half full interrupt
		 * in the data out loop, but that would be overkill, considering the intended
		 * use of this ioctl
		 */

		if (count > 255) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at DATA_OUT - too many bytes!\n",
			       instance);
			return -EINVAL;
		}

		if (copy_from_user(data_array, (caddr_t) arg, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_from_user error!\n",
			       instance);
			return -EINVAL;
		}

		for (i = 0; i < count; i++) {
			if (
			    (IHCP_GETL(IHCP_INTERFACE_STATUS) &
			     FIFO_NOT_FULL) == 0) {
				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at DATA_OUT - fifo full! 0x%x bytes requested, 0x%x bytes transferred\n",
				       instance, count, i);
				return -EIO;
			}

			lt = data_array[i];
			if (unit_p->byte_order == HIGH_BYTE_FIRST)
				lt = lt << 24;
			IHCP_PUTL(IHCP_8_BIT_DATA_OUT, lt);
		}

		break;


	case MASK & IHCPIO_RDY_WAIT:	/* wait for rdy & fifo empty - or timeout */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at RDY_WAIT, wait time = %ld (jiffies)\n",
			instance, unit_p->fifo_time));

		/*
		 * if already ready and fifo empty, just return
		 * first clear any stale flags
		 */

		unit_p->unit_flags &= IHCP_CLEAR_FLAGS;

		if (IHCP_GETL(IHCP_INTERFACE_STATUS) &
		    DEVICE_AND_INT_READY) return 0;

		/*
		 * make sure no left over interrupt flag
		 */

		IHCP_PUTL(IHCP_INTERFACE_CONTROL, CLEAR_INTERRUPT_FLAG);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

		/*
		 * save processor flags, disable interrupts, and take spinlock w/mutex
		 * BEFORE ENABLING BOARD INTERRUPTS!
		 */

		mutex_lock(&unit_p->mutex);

		/*
		 * set sleeping flag, so we can detect an abnormal return from sleep
		 */

		unit_p->sleeping = 1;

		/*
		 * set flags to indicate for what we wait
		 */

		unit_p->unit_flags |= IHCP_RDY_WAIT;

		/*
		 * enable plx and ihcp interrupts for device and interface ready
		 * we don't or into ihcp or plx mask reg since we know exactly which ints to enb
		 */

		IHCP_PUTL(IHCP_INTERRUPT_MASK, DEV_AND_INT_RDY_INT_ENB);
		PLX_PUTL(PLX_INT_CSTAT,
			 (PCI_INTERRUPT_ENABLE | PCI_LOCAL_INT_ENABLE));

		/*
		 * snooze until ready interrupt, timeout, or signal
		 */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at RDY_WAIT, time_now = %ld, unit_p->fifo_time = %ld\n",
			instance, jiffies, unit_p->fifo_time));

		wait_return =
		    cond_timed_wait_rel(&unit_p->cond, &unit_p->mutex,
					unit_p->fifo_time);

		/*
		 * make sure interrupt enables are off
		 * and clear interrupt flag in case we got here via timeout just
		 * as interrupt set the flag
		 */

		IHCP_PUTL(IHCP_INTERRUPT_MASK, 0);
		PLX_PUTL(PLX_INT_CSTAT, 0);

		IHCP_PUTL(IHCP_INTERFACE_CONTROL, CLEAR_INTERRUPT_FLAG);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

		/*
		 * re-enable interrupts (if previously on)
		 * and release spin lock via mutex
		 */

		mutex_unlock(&unit_p->mutex);

		/*
		 * find out why we woke up (we could probably dump if sleeping and use the return
		 * value of cond_timed_wait_rel directly, but ...
		 */

		if (unit_p->sleeping) {

			unit_p->sleeping = 0;

			DPRINT((KERN_DEBUG
				"ihcp_ioctl: instance %d: at RDY_WAIT, abnormal return from cond_timed_wait_rel()!\n",
				instance));

			/*
			 * check for timeout
			 */

			if (wait_return == COND_WAIT_TIMEOUT) {

				/*
				 * timeout may not be an error, so no error message just flag and return
				 * caller can test for timeout on error return
				 */

				DPRINT((KERN_DEBUG
					"ihcp_ioctl: instance %d: at RDY_WAIT, timeout while waiting for interrupt!\n",
					instance));
				unit_p->unit_flags |= IHCP_RDY_TIMEOUT;
				return -EIO;
			}

			/*
			 * check for signal
			 */

			if (wait_return == COND_WAIT_SIGNAL) {

				/*
				 * flag signal received and print error message
				 */

				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at RDY_WAIT, signal received while waiting for interrupt\n",
				       instance);
				unit_p->unit_flags |= IHCP_SIG_RECEIVED;
				return -EINTR;
			}

		}
		/* end of if sleeping       */
		return 0;

		break;


	case MASK & IHCPIO_HALF_WAIT:	/* wait for fifo <half full int-or timeout */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at HALF_WAIT, wait time = %ld (jiffies)\n",
			instance, unit_p->fifo_time));

		/*
		 * if already less than half full, just return
		 * first clear any stale flags
		 */

		unit_p->unit_flags &= IHCP_CLEAR_FLAGS;

		if (IHCP_GETL(IHCP_INTERFACE_STATUS) & FIFO_NOT_HALF_FULL)
			return 0;

		/*
		 * clear any left over interrupt flag
		 */

		IHCP_PUTL(IHCP_INTERFACE_CONTROL, CLEAR_INTERRUPT_FLAG);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

		/*
		 * save processor flags, disable interrupts, and take spinlock w/mutex
		 * BEFORE ENABLING BOARD INTERRUPTS!
		 */

		mutex_lock(&unit_p->mutex);
		/*
		 * set sleeping flag, so we can detect an abnormal return from sleep
		 */

		unit_p->sleeping = 1;

		/*
		 * set flags to indicate for what we wait
		 */

		unit_p->unit_flags |= IHCP_HALF_WAIT;

		/*
		 * enable plx and ihcp interrupt for fifo less than half full
		 * we don't or into ihcp or plx mask reg since we know exactly which ints to enb
		 */

		IHCP_PUTL(IHCP_INTERRUPT_MASK, FIFO_NOT_HALF_INT_ENB);
		PLX_PUTL(PLX_INT_CSTAT,
			 (PCI_INTERRUPT_ENABLE | PCI_LOCAL_INT_ENABLE));

		/*
		 * snooze until ready interrupt, timeout, or signal
		 */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at HALF_WAIT, time_now = %ld, unit_p->fifo_time = %ld\n",
			instance, jiffies, unit_p->fifo_time));

		wait_return =
		    cond_timed_wait_rel(&unit_p->cond, &unit_p->mutex,
					unit_p->fifo_time);

		/*
		 * make sure interrupt enables are off and clear flag in case we
		 * got here via timeout just as interrupt happened
		 */

		IHCP_PUTL(IHCP_INTERRUPT_MASK, 0);
		PLX_PUTL(PLX_INT_CSTAT, 0);

		IHCP_PUTL(IHCP_INTERFACE_CONTROL, CLEAR_INTERRUPT_FLAG);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

		/*
		 * re-enable interrupts (if previously on)
		 * and release spin lock via mutex
		 */

		mutex_unlock(&unit_p->mutex);

		/*
		 * find out why we woke up (we could probably dump if sleeping and use the return
		 * value of cond_timed_wait_rel directly, but ...
		 */

		if (unit_p->sleeping) {

			unit_p->sleeping = 0;

			DPRINT((KERN_DEBUG
				"ihcp_ioctl: instance %d: at HALF_WAIT, abnormal return from cond_timed_wait_rel()!\n",
				instance));

			/*
			 * check for timeout
			 */

			if (wait_return == COND_WAIT_TIMEOUT) {

				/*
				 * timeout may not be an error, so no error message just flag and return
				 * caller can test for timeout on error return
				 */

				DPRINT((KERN_DEBUG
					"ihcp_ioctl: instance %d: at HALF_WAIT, timeout while waiting for interrupt!\n",
					instance));
				unit_p->unit_flags |= IHCP_HALF_TIMEOUT;
				return -EIO;
			}

			/*
			 * check for signal
			 */

			if (wait_return == COND_WAIT_SIGNAL) {

				/*
				 * flag signal received and print error message
				 */

				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at HALF_WAIT, signal received while waiting for interrupt\n",
				       instance);
				unit_p->unit_flags |= IHCP_SIG_RECEIVED;
				return -EINTR;
			}

		}
		/* end of if sleeping       */
		return 0;

		break;

	case MASK & IHCPIO_STREAM_ON:	/* set streaming mode if 10106 */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at STREAM_ON\n",
			instance));

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) != IHCPIO_CENT) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at STREAM_ON - attempt to set streaming mode on board strapped for versatec!\n",
			       instance);
			return -ENOTTY;
		}

		/*
		 * this command talks to the fifo - make sure there is room !!
		 */

		if ((IHCP_GETL(IHCP_INTERFACE_STATUS) & FIFO_NOT_FULL) ==
		    0) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at STREAM_ON - no room in fifo for stream on command!\n",
			       instance);
			return -EIO;
		}

		lt = SET_IO_MODE | DATA_STREAMING_MODE;

		if (unit_p->byte_order == HIGH_BYTE_FIRST)
			lt = lt << 24;
		IHCP_PUTL(IHCP_COMMAND_OUT, lt);

		/*
		 * save streaming mode for compatibility with old vme ioctls
		 */

		unit_p->data_stream_mode = DATA_STREAMING_ON;

		break;

	case MASK & IHCPIO_STREAM_OFF:	/* turn streaming off (if 10106) */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at STREAM_OFF\n",
			instance));

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) != IHCPIO_CENT) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at STREAM_OFF - attempt to clear streaming mode on board strapped for versatec!\n",
			       instance);
			return -ENOTTY;
		}

		/*
		 * this command talks to the fifo - make sure there is room !!
		 */

		if ((IHCP_GETL(IHCP_INTERFACE_STATUS) & FIFO_NOT_FULL) ==
		    0) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at STREAM_OFF - no room in fifo for stream off command!\n",
			       instance);
			return -EIO;
		}

		/*
		 * set_io_mode w/ nothing else or'd in turns off streaming
		 * and auto ltr - we won't try to do streaming and auto ltr
		 * at same time in this driver
		 */

		lt = SET_IO_MODE;

		if (unit_p->byte_order == HIGH_BYTE_FIRST)
			lt = lt << 24;
		IHCP_PUTL(IHCP_COMMAND_OUT, lt);

		/*
		 * save streaming mode for compatibility with old vme ioctls
		 */

		unit_p->data_stream_mode = DATA_STREAMING_OFF;

		break;

	case MASK & IHCPIO_GET_FLAGS:	/* get unit flag longword */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at GET_FLAGS\n",
			instance));

		if (copy_to_user
		    ((caddr_t) arg, &unit_p->unit_flags, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_to_user error!\n",
			       instance);
			return -EINVAL;
		}

		break;

	case MASK & IHCPIO_GET_FIFO:	/* get formatted fifo flags */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at GET_FIFO\n",
			instance));

		/*
		 * get fifo status bits and swizzle and complement
		 * to match sbus driver format
		 */

		lt = IHCP_GETL(IHCP_INTERFACE_STATUS);

		bits = 0;
		if (!(lt & FIFO_NOT_FULL))
			bits |= IHCPIO_FIFO_FULL;
		if (!(lt & FIFO_NOT_HALF_FULL))
			bits |= IHCPIO_FIFO_HALF;
		if (lt & FIFO_EMPTY)
			bits |= IHCPIO_FIFO_EMPTY;

		if (copy_to_user((caddr_t) arg, &bits, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_to_user error!\n",
			       instance);
			return -EINVAL;
		}

		break;

	case MASK & IHCPIO_MASTER_CLEAR:	/* master clear board - not for use by wimps!   */
						/* clear ihcp logic and restore all defaults    */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at MASTER_CLEAR\n",
			instance));

		IHCP_PUTL(IHCP_INTERFACE_CONTROL, MASTER_CLEAR);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

		/*
		 * make absolutely sure that dma and interrupts are off in plx chip
		 */

		PLX_PUTL(PLX_INT_CSTAT, 0);
		PLX_PUTL(PLX_DMA_CMD_STAT_BOTH, 0);

		/*
		 * set and clear mclr bit - other bits in this register are also
		 * treated like pulses, and do not persist, so we don't need to
		 * or in a bit, then and it out
		 */

		IHCP_PUTL(IHCP_INTERFACE_CONTROL, MASTER_CLEAR);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

		/*
		 * set and clear the device reset bit, with a delay and write
		 * buffer flush to guarantee the delay
		 * 
		 * the mode read between puts should force the write buffer to flush
		 */

		IHCP_PUTL(IHCP_INTERFACE_CONTROL,
			  (IHCP_GETL(IHCP_INTERFACE_CONTROL) |
			   RESET_DEVICE));
		read_dump_0 = IHCP_GETL(IHCP_MODE);
		udelay(1);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL,
			  (IHCP_GETL(IHCP_INTERFACE_CONTROL) &
			   ~RESET_DEVICE));

		/*
		 * set to default print mode, default handshake speed, and default
		 * byte ordering
		 * 
		 * we don't have to check for fifo space, since we just cleared it
		 * 
		 * print mode is set only if the board is strapped for versatec
		 * 
		 */

		if (ihcp_byte_order_def == HIGH_BYTE_FIRST) {
			IHCP_PUTL(IHCP_MODE,
				  (IHCP_GETL(IHCP_MODE) | BYTE_SWIZZLE));
			PLX_PUTL(PLX_EEPROM_USER,
				 (PLX_GETL(PLX_EEPROM_USER) &
				  ~USER_OUTPUT));
			unit_p->byte_order = ihcp_byte_order_def;
		}


		if ((unit_p->unit_flags & IHCP_BOARD_MASK) != IHCPIO_CENT) {

			/*
			 * versatec - set speed and print/plot mode
			 */

			IHCP_PUTL(IHCP_MODE,
				  ((IHCP_GETL(IHCP_MODE) & ~SPEED_MASK) |
				   unit_p->vers_speed_def));

			lt = SET_VERSATEC_MODE | unit_p->mode_def;
			if (unit_p->byte_order == HIGH_BYTE_FIRST)
				lt = lt << 24;
			IHCP_PUTL(IHCP_COMMAND_OUT, lt);

			unit_p->mode = unit_p->mode_def;

		} else {

			/*
			 * centronics - set speed only
			 */


			IHCP_PUTL(IHCP_MODE,
				  ((IHCP_GETL(IHCP_MODE) & ~SPEED_MASK) |
				   unit_p->cent_speed_def));


			/*
			 * clear saved data streaming mode
			 */

			unit_p->data_stream_mode = DATA_STREAMING_OFF;
		}

		break;

	case MASK & IHCPIO_SOFT_ACK:	/* software ack - looks like an ack from the dev */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at SOFT_ACK\n",
			instance));

		/*
		 * set and clear soft ack bit, as in mclr
		 */

		IHCP_PUTL(IHCP_INTERFACE_CONTROL, SOFTWARE_ACK);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

		break;

	case MASK & IHCPIO_AUTO_LTR_COUNT:	/* set byte count for auto ltr function */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at AUTO_LTR_COUNT, byte count = %ld\n",
			instance, bits));

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) == IHCPIO_CENT) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at AUTO_LTR_COUNT - attempt to set auto ltr count on board strapped for centronics!\n",
			       instance);
			return -ENOTTY;
		}

		/*
		 * auto ltr count is 16 bit register, accessed as two
		 * separate bytes
		 * 
		 * the count register takes count minus 1
		 */

		IHCP_PUTL(IHCP_AUTO_LTR_LOW, (bits - 1) & 0xFF);
		IHCP_PUTL(IHCP_AUTO_LTR_HIGH, ((bits - 1) >> 8) & 0xFF);

		break;

	case MASK & IHCPIO_AUTO_LTR_ON:	/* enable auto ltr operation    */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at AUTO_LTR_ON\n",
			instance));

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) == IHCPIO_CENT) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at AUTO_LTR_ON - attempt to set auto ltr mode on board strapped for centronics!\n",
			       instance);
			return -ENOTTY;
		}

		/*
		 * this command talks to the fifo - make sure there is room !!
		 */

		if ((IHCP_GETL(IHCP_INTERFACE_STATUS) & FIFO_NOT_FULL) ==
		    0) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at AUTO_LTR_ON - no room in fifo for auto ltr on command!\n",
			       instance);
			return -EIO;
		}

		lt = SET_IO_MODE | AUTO_LTR_MODE;

		if (unit_p->byte_order == HIGH_BYTE_FIRST)
			lt = lt << 24;
		IHCP_PUTL(IHCP_COMMAND_OUT, lt);

		break;

	case MASK & IHCPIO_AUTO_LTR_OFF:	/* disable auto ltr operation   */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at AUTO_LTR_OFF\n",
			instance));

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) == IHCPIO_CENT) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at AUTO_LTR_OFF - attempt to clear auto ltr mode on board strapped for centronics!\n",
			       instance);
			return -ENOTTY;
		}

		/*
		 * this command talks to the fifo - make sure there is room !!
		 */

		if ((IHCP_GETL(IHCP_INTERFACE_STATUS) & FIFO_NOT_FULL) ==
		    0) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at AUTO_LTR_OFF - no room in fifo for auto ltr off command!\n",
			       instance);
			return -EIO;
		}

		/*
		 * set io mode with no other bits clears auto ltr and data streaming
		 */

		lt = SET_IO_MODE;

		if (unit_p->byte_order == HIGH_BYTE_FIRST)
			lt = lt << 24;
		IHCP_PUTL(IHCP_COMMAND_OUT, lt);

		break;

	case MASK & IHCPIO_DEV_AND_VEND_ID:	/* return vendor id in low 16 bits, device id in high 16 bits   */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at DEV_AND_VEND_ID\n",
			instance));

		bits = unit_p->dev_and_vendor_id;

		if (copy_to_user((caddr_t) arg, &bits, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_to_user error!\n",
			       instance);
			return -EINVAL;
		}

		break;

	case MASK & IHCPIO_REVISION_ID:	/* return board revision id     */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at REVISION_ID\n",
			instance));

		bits = unit_p->revision_id;

		if (copy_to_user((caddr_t) arg, &bits, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_to_user error!\n",
			       instance);
			return -EINVAL;
		}

		break;

	case MASK & IHCPIO_LITTLE_ENDIAN:	/* data will go to plotter low byte first       */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at LITTLE_ENDIAN\n",
			instance));

		/*
		 * low first byte ordering (the hardware default) is set by turning off
		 * the byte swizzle bit in the ihcp logic, and TURNING ON
		 * the user output bit in the plx logic]
		 */

		IHCP_PUTL(IHCP_MODE,
			  (IHCP_GETL(IHCP_MODE) & ~BYTE_SWIZZLE));
		PLX_PUTL(PLX_EEPROM_USER,
			 (PLX_GETL(PLX_EEPROM_USER) | USER_OUTPUT));

		unit_p->byte_order = LOW_BYTE_FIRST;

		break;

	case MASK & IHCPIO_BIG_ENDIAN:	/* data will go to plotter high byte first      */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at BIG_ENDIAN\n",
			instance));

		/*
		 * turn on the ihcp byte swizzle bit and TURN OFF the plx
		 * user output bit
		 */

		IHCP_PUTL(IHCP_MODE,
			  (IHCP_GETL(IHCP_MODE) | BYTE_SWIZZLE));
		PLX_PUTL(PLX_EEPROM_USER,
			 (PLX_GETL(PLX_EEPROM_USER) & ~USER_OUTPUT));

		unit_p->byte_order = HIGH_BYTE_FIRST;

		break;

	case MASK & IHCPIO_DIRECT_MODE:	/* direct user write of mode register - use carefully!  */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at DIRECT_MODE\n",
			instance));

		IHCP_PUTL(IHCP_MODE, bits);

		break;

	case MASK & IHCPIO_DEVICE_CONTROL:	/* direct write to device control register      */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at DEVICE_CONTROL\n",
			instance));

		IHCP_PUTL(IHCP_DEVICE_CONTROL, bits);

		break;

	case MASK & IHCPIO_INTERFACE_STATUS:	/* direct read of interface status register     */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at INTERFACE_STATUS\n",
			instance));

		bits = IHCP_GETL(IHCP_INTERFACE_STATUS);

		if (copy_to_user((caddr_t) arg, &bits, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_to_user error!\n",
			       instance);
			return -EINVAL;
		}

		break;

	case MASK & IHCPIO_DEVICE_STATUS:	/* direct read of device status register        */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at DEVICE_STATUS\n",
			instance));

		bits = IHCP_GETL(IHCP_DEVICE_STATUS);

		if (copy_to_user((caddr_t) arg, &bits, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_to_user error!\n",
			       instance);
			return -EINVAL;
		}

		break;

	case MASK & IHCPIO_REVERSE_DATA:	/* direct read of reverse data register - requires      */
						/* fiddling of tri-state bit, and others to actually    */
						/* do anything useful with this                         */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: istance %d: at REVERSE_DATA\n",
			instance));

		bits = IHCP_GETL(IHCP_REVERSE_DATA);

		if (copy_to_user((caddr_t) arg, &bits, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_to_user error!\n",
			       instance);
			return -EINVAL;
		}

		break;

		/*
		 * ioctls for compatibility with versatec and sun drivers for VME and Sbus boards
		 */

	case MASK & LPSETVERSATEC:	/* compatibility - set versatec port */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at LPSETVERSATEC\n",
			instance));

		/*
		 * make sure the board is a versatec type - can't do this w/centr
		 */

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) == IHCPIO_CENT) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at LPSETVERSATEC - attempt to select versatec port onn board strapped for centronics!\n",
			       instance);
			return -ENOTTY;
		}

		break;

	case MASK & LPSETCENTRONICS:	/* compatibility - set centr port */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at LPSETCENTRONICS\n",
			instance));

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) != IHCPIO_CENT) {
			printk(KERN_WARNING
			       "ihcp_ioctl: instance %d: at LPSETCENTRONICS - attempt to select centronics port on board strapped for versatec!\n",
			       instance);
			return -ENOTTY;
		}

		break;


	case MASK & LPCOMMAND:	/* compatibility - old style cmds */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at LPCOMMAND\n",
			instance));

		/*
		 * as much as possible, allow multiple command bits in the argument.
		 * this is probably not the way it was intended, but we want to
		 * protect ourselves against non-reasonable uses of the old calls
		 */

		/*
		 * assert long reset signal
		 */

		if (bits & LPC_LRST)
			IHCP_PUTL(IHCP_INTERFACE_CONTROL,
				  IHCP_GETL(IHCP_INTERFACE_CONTROL) |
				  RESET_DEVICE);

		/*
		 * de-assert long reset
		 */

		if (bits & LPC_DRST)
			IHCP_PUTL(IHCP_INTERFACE_CONTROL,
				  IHCP_GETL(IHCP_INTERFACE_CONTROL) &
				  ~RESET_DEVICE);;

		/*
		 * test for option port request - better be strapped for centronics!!
		 */

		if (bits & LPC_SOPT)

			if ((unit_p->unit_flags & IHCP_BOARD_MASK) !=
			    IHCPIO_CENT) {
				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at LPCOMMAND - attempt to select centronics port on board strapped for versatec!\n",
				       instance);
				return -ENOTTY;
			}


		/*
		 * test for versatec port selection
		 */

		if (bits & LPC_SVPT)

			if ((unit_p->unit_flags & IHCP_BOARD_MASK) ==
			    IHCPIO_CENT) {
				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at LPCOMMAND - attempt to select versatec port on board strapped for centronics!\n",
				       instance);
				return -ENOTTY;
			}


		/*
		 * enable data streaming - if board is strapped for centronics!
		 */

		if (bits & LPC_DSTR) {

			if ((unit_p->unit_flags & IHCP_BOARD_MASK) !=
			    IHCPIO_CENT) {
				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at LPCOMMAND - attempt to set streaming mode on board strapped for versatec!\n",
				       instance);
				return -ENOTTY;
			}

			/*
			 * this command talks to the fifo - make sure there is room !!
			 */

			if (
			    (IHCP_GETL(IHCP_INTERFACE_STATUS) &
			     FIFO_NOT_FULL) == 0) {
				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at LPCOMMAND - no room in fifo for stream on command!\n",
				       instance);
				return -EIO;
			}

			lt = SET_IO_MODE | DATA_STREAMING_MODE;

			if (unit_p->byte_order == HIGH_BYTE_FIRST)
				lt = lt << 24;
			IHCP_PUTL(IHCP_COMMAND_OUT, lt);

			/*
			 * save for posterity
			 */

			unit_p->data_stream_mode = DATA_STREAMING_ON;
		}


		/* reset data streaming bit */

		if (bits & LPC_DDST) {

			if ((unit_p->unit_flags & IHCP_BOARD_MASK) !=
			    IHCPIO_CENT) {
				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at LPCOMMAND - attempt to clear streaming mode on non-centronics unit!\n",
				       instance);
				return -ENOTTY;
			}

			if (
			    (IHCP_GETL(IHCP_INTERFACE_STATUS) &
			     FIFO_NOT_FULL) == 0) {
				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at LPCOMMAND - no room in fifo for stream off command!\n",
				       instance);
				return -EIO;
			}

			/*
			 * set_io_mode w/ nothing else or'd in turns off streaming
			 * and auto ltr - we won't try to do streaming and auto ltr
			 * at same time in this driver
			 */

			lt = SET_IO_MODE;

			if (unit_p->byte_order == HIGH_BYTE_FIRST)
				lt = lt << 24;
			IHCP_PUTL(IHCP_COMMAND_OUT, lt);

			/*
			 * save streaming mode for later getregs call
			 */

			unit_p->data_stream_mode = DATA_STREAMING_OFF;
		}

		/*
		 * test for versatec print/plot mode changes - make sure we are
		 * a versatec board !!
		 */

		if (bits & LPC_PRINTMASK) {

			if ((unit_p->unit_flags & IHCP_BOARD_MASK) ==
			    IHCPIO_CENT) {
				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at LPCOMMAND - attempt to do versatec mode set on board strapped for centronics!\n",
				       instance);
				return -ENOTTY;
			}

			/*
			 * this cmd talks to the fifo - make sure there is room !!
			 */

			if (
			    (IHCP_GETL(IHCP_INTERFACE_STATUS) &
			     FIFO_NOT_FULL) == 0) {
				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at LPCOMMAND - no room in fifo for versatec mode set!\n",
				       instance);
				return -EIO;
			}

			/*
			 * get current mode setting
			 */

			lt = unit_p->mode;

			/* now look at individual command bits and twiddle mode */

			if (bits & LPC_VSPP)
				lt |= SPP_MODE;	/* set spp mode */
			if (bits & LPC_DSPP)
				lt &= ~SPP_MODE;	/* clear spp   */
			if (bits & LPC_VPLT)
				lt |= PLOT_MODE;	/* set plot mode */
			if (bits & LPC_DVPT)
				lt &= ~PLOT_MODE;	/* clear plot */
			if (bits & LPC_PRNT)
				lt = NORM_PRINT_MODE;	/* 0 = normal print mode     */

			/*
			 * save final mode - without mode set command bit - 
			 * in unit structure
			 */

			unit_p->mode = lt;

			/*
			 * write final mode to fifo - remember to write to high
			 * byte if big-endian data mode selected
			 */

			lt |= SET_VERSATEC_MODE;
			if (unit_p->byte_order == HIGH_BYTE_FIRST)
				lt = lt << 24;
			IHCP_PUTL(IHCP_COMMAND_OUT, lt);

		}


		/*
		 * check for software ack pulse request
		 */

		if (bits & LPC_SACK) {

			/*
			 * set and clear soft ack bit, as in mclr
			 */

			IHCP_PUTL(IHCP_INTERFACE_CONTROL, SOFTWARE_ACK);
			IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);
		}

		/*
		 * check for master clear request - the old vme board also issued a reset
		 * to the device when mclr was pulsed, so we will do that also
		 * 
		 * we will also restore the default print/plot mode (if versatec),
		 * the device handshake speed, and the byte ordering
		 */

		if (bits & LPC_MCLR) {

			IHCP_PUTL(IHCP_INTERFACE_CONTROL, MASTER_CLEAR);
			IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

			/*
			 * make absolutely sure that dma and interrupts are off in plx chip
			 */

			PLX_PUTL(PLX_INT_CSTAT, 0);
			PLX_PUTL(PLX_DMA_CMD_STAT_BOTH, 0);

			/*
			 * set and clear mclr bit - other bits in this register are also
			 * treated like pulses, and do not persist, so we don't need to
			 * or in a bit, then and it out
			 */

			IHCP_PUTL(IHCP_INTERFACE_CONTROL, MASTER_CLEAR);
			IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

			/*
			 * set and clear the device reset bit, with a delay and write
			 * buffer flush to guarantee the delay
			 * 
			 * the mode read between puts should force the write buffer to flush
			 */

			IHCP_PUTL(IHCP_INTERFACE_CONTROL,
				  (IHCP_GETL(IHCP_INTERFACE_CONTROL) |
				   RESET_DEVICE));
			read_dump_0 = IHCP_GETL(IHCP_MODE);
			udelay(1);
			IHCP_PUTL(IHCP_INTERFACE_CONTROL,
				  (IHCP_GETL(IHCP_INTERFACE_CONTROL) &
				   ~RESET_DEVICE));

			/*
			 * set to default print mode, default handshake speed, and default
			 * byte ordering
			 * 
			 * we don't have to check for fifo space, since we just cleared it
			 * 
			 * print mode is set only if the board is strapped for versatec
			 * 
			 */

			if (ihcp_byte_order_def == HIGH_BYTE_FIRST) {
				IHCP_PUTL(IHCP_MODE,
					  (IHCP_GETL(IHCP_MODE) |
					   BYTE_SWIZZLE));
				PLX_PUTL(PLX_EEPROM_USER,
					 (PLX_GETL(PLX_EEPROM_USER) &
					  ~USER_OUTPUT));
				unit_p->byte_order = ihcp_byte_order_def;
			}


			if ((unit_p->unit_flags & IHCP_BOARD_MASK) !=
			    IHCPIO_CENT) {

				/*
				 * versatec - set speed and print/plot mode
				 */

				IHCP_PUTL(IHCP_MODE,
					  ((IHCP_GETL
					    (IHCP_MODE) & ~SPEED_MASK) |
					   unit_p->vers_speed_def));

				lt = SET_VERSATEC_MODE | unit_p->mode_def;
				if (unit_p->byte_order == HIGH_BYTE_FIRST)
					lt = lt << 24;
				IHCP_PUTL(IHCP_COMMAND_OUT, lt);

				unit_p->mode = unit_p->mode_def;

			} else {

				/*
				 * centronics - set speed only
				 */


				IHCP_PUTL(IHCP_MODE,
					  ((IHCP_GETL
					    (IHCP_MODE) & ~SPEED_MASK) |
					   unit_p->cent_speed_def));
			}

			/*
			 * clear saved data streaming mode
			 */

			unit_p->data_stream_mode = DATA_STREAMING_OFF;
		}

		/*
		 * look for pulsed command request that has to go through fifo
		 */

		if (bits & (LPC_VCLR | LPC_VTFF | LPC_VEOT | LPC_VLTR)) {

			/* make sure the board is a versatec type - can't do w/centr */

			if ((unit_p->unit_flags & IHCP_BOARD_MASK) ==
			    IHCPIO_CENT) {
				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at LPCOMMAND - attempt to do versatec pulsed command to board strapped for centronics!\n",
				       instance);
				return -ENOTTY;
			}

			/*
			 * this cmd talks to the fifo - make sure there is room !! 
			 * WE WILL ASSUME THAT THERE WILL ONLY BE ONE PULSE PER CALL
			 */

			if (
			    (IHCP_GETL(IHCP_INTERFACE_STATUS) &
			     FIFO_NOT_FULL) == 0) {
				printk(KERN_WARNING
				       "ihcp_ioctl: instance %d: at LPCOMMAND - no room in fifo for versatec pulsed command!\n",
				       instance);
				return -EIO;
			}

			/*
			 * shuffle vme style commands to match pci board
			 */

			if (bits & LPC_VLTR)
				lt = VERSATEC_LINE_TERM;
			if (bits & LPC_VEOT)
				lt = VERSATEC_EOT;
			if (bits & LPC_VTFF)
				lt = VERSATEC_FORM_FEED;
			if (bits & LPC_VCLR)
				lt = VERSATEC_CLEAR;

			if (unit_p->byte_order == HIGH_BYTE_FIRST)
				lt = lt << 24;
			IHCP_PUTL(IHCP_COMMAND_OUT, lt);
		}

		break;

	case MASK & LPGETREGS:	/* old style (VMEbus) get regs command  */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at LPGETREGS\n",
			instance));
		/*
		 * start work on interface status register
		 */

		st = 0;

		/*
		 * get interface status register
		 * and fiddle interface and device ready bits
		 */

		ct = (u_char) IHCP_GETL(IHCP_INTERFACE_STATUS);

		if (ct & DEVICE_AND_INT_READY)
			st |= OLD_DIRY;
		if (ct & DEVICE_READY)
			st |= OLD_DVRY;

		/*
		 * get remembered data streaming mode
		 */

		if (unit_p->data_stream_mode == DATA_STREAMING_ON)
			st |= OLD_DSTR;

		/*
		 * get board type from unit struct
		 * if centronics, flag as old style option port
		 */

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) == IHCPIO_CENT)
			st |= OLD_SOPT;

		/*
		 * save result in cis portion of lp_regs structure
		 * 
		 * we used to put it in the top of a u_long, and the cds portion
		 * in the bottom, but we get into trouble with "endian-ness"
		 * moving across the various platforms
		 * 
		 * starting with this port, we will use the struct instead
		 */

		lp_regs.cisreg = st;

		/*
		 * do the same ugly things with the device status register -- 
		 * it has two halves: versatec and centronics -- only
		 * work on the appropriate half
		 */

		st = 0;

		if ((unit_p->unit_flags & IHCP_BOARD_MASK) == IHCPIO_CENT) {

			/*
			 * work on centronics part
			 * get the device status register and
			 * shuffle bit positions and polarities
			 */

			ct = (u_char) IHCP_GETL(IHCP_DEVICE_STATUS);

			if (ct & VERSATEC_READY)
				st |= OLD_OACK;
			if (!(ct & CENTRONICS_BUSY))
				st |= OLD_ONBY;
			if (!(ct & PAPER_OUT))
				st |= OLD_OPPR;
			if (!(ct & ONLINE))
				st |= OLD_ONSL;
			if (ct & FAULT)
				st |= OLD_OFLT;

			/*
			 * look for device reset latch
			 */

			if (IHCP_GETL(IHCP_INTERFACE_CONTROL) &
			    RESET_DEVICE) st |= OLD_OLRS;	/* long reset     */
		} else {

			/*
			 * work on versatec part
			 */

			if ((unit_p->unit_flags & IHCP_BOARD_MASK) ==
			    IHCPIO_VERS_TTL) st |= OLD_VTTL;

			/*
			 * get device status and twiddle bits
			 */

			ct = (u_char) IHCP_GETL(IHCP_DEVICE_STATUS);

			if (ct & VERSATEC_READY)
				st |= OLD_VRDY;	/* vers ready */
			if (ct & PAPER_OUT)
				st |= OLD_VPPR;	/* paper ok */
			if (ct & ONLINE)
				st |= OLD_VONL;	/* vers onlin */

			/*
			 * get current mode from unit structure
			 */

			ct = unit_p->mode;

			if (ct & SPP_MODE)
				st |= OLD_VSPP;	/* spp mode */
			if (ct & PLOT_MODE)
				st |= OLD_VPLT;	/* plot mode */
		}

		/*
		 * save in cds portion of status structure
		 */

		lp_regs.cdsreg = st;

		if (copy_to_user((caddr_t) arg, &lp_regs, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_to_user error!\n",
			       instance);
			return -EINVAL;
		}

		break;

	case MASK & LPSETTIMVAL:	/* set timeout value - old style - for dma and fifo empty       */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at LPSETTIMVAL\n",
			instance));

		/*
		 * old style timeout call used fiftieths of a second as argument
		 */

		if (bits < DMA_TIME_MIN * 50)
			bits = DMA_TIME_MIN * 50;
		if (bits > DMA_TIME_MAX * 50)
			bits = DMA_TIME_MAX * 50;

		unit_p->fifo_time = (u_long) ((bits * HZ) / 50);
		unit_p->dma_time = (u_long) ((bits * HZ) / 50);

		break;


	case MASK & LPGETTIMVAL:	/* get current timeout (in ticks ?)     */

		DPRINT((KERN_DEBUG
			"ihcp_ioctl: instance %d: at LPGETTIMVAL\n",
			instance));

		/*
		 * return # of fiftieths of a second
		 */

		lt = (unit_p->dma_time / HZ) * 50;

		if (copy_to_user((caddr_t) arg, &lt, count)) {
			printk(KERN_ERR
			       "ihcp_ioctl: instance %d: copy_to_user error!\n",
			       instance);
			return -EINVAL;
		}

		break;


	default:

		/*
		 * we don't print an error message here
		 * if we are involved in a pipe, the pipe logic may send
		 * us a generic ioctl to determine if we are a terminal
		 * we have to return an error or we will give a false
		 * terminal indication.
		 */

		return -EINVAL;

		break;
	}

	return 0;
}



/*
 * our interrupt handler - registered with the system in init_module
 * each board's instance will register the same handler, but with a different unit structure pointer
 * when called, dev_id is a pointer at the soft state structure for this instance
 * 
 * as of the 2.3.x kiobuf version, we use mutex and conditional variable for SMP compatibility
 */

static void ihcp_intr(int irq, void *dev_id, struct pt_regs *regs)
{

	struct ihcp_unit_t	*unit_p;	/* will point at our soft state                 */
	volatile u_long		temp;		/* just what it says                            */
	int			instance;	/* this board's minor #                         */
	int			wake_up_needed;	/* 1 if wake up needed                          */


	/*
	 * get the unit struct pointer this way to avoid parameter type complaints by the compiler
	 * get instance for debugging
	 * clear wake up needed flag
	 */

	unit_p = (struct ihcp_unit_t *) dev_id;
	instance = unit_p->instance;
	wake_up_needed = 0;

	DPRINT((KERN_DEBUG
		"ihcp_intr: instance %d: entering ihcp_intr (may not be our interrupt)\n",
		instance));

	/*
	 * grab mutex - in case strategy and int thread are running on different cpu's ???
	 */

	mutex_lock(&unit_p->mutex);

	/*
	 * get plx interrupt control/status register and check for interrupt
	 * 
	 * check for master interrupt enable bit and either local int rq or dma int rq
	 */


	temp = PLX_GETL(PLX_INT_CSTAT);

	if ((temp & PCI_INTERRUPT_ENABLE)
	    && ((temp & LOCAL_INTERRUPT) || (temp & DMA_0_INTERRUPT))) {

		DPRINT((KERN_DEBUG
			"ihcp_intr: instance %d: interrupt claimed, unit_p = 0x%p\n",
			instance, unit_p));

		/*
		 * force off dma enable, interrupt enables, and ihcp interrupt mask bits
		 * which should release the interrupt request
		 * 
		 * then (after masks are off) toggle reset int flag bit on & off to clear
		 * interrupt flag (which is a latch)
		 * and do the same to the dma done interrupt
		 */

		PLX_PUTL(PLX_DMA_CMD_STAT_BOTH, 0);
		PLX_PUTL(PLX_INT_CSTAT, 0);
		IHCP_PUTL(IHCP_INTERRUPT_MASK, 0);

		PLX_PUTL(PLX_DMA_CMD_STAT_BOTH, DMA_CLEAR_INTERRUPT);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, CLEAR_INTERRUPT_FLAG);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

		/*
		 * do register reads to get the previous writes to flush to the pci bus
		 * so the writes that clear the interrupt will be complete before we
		 * exit - we will hope that the optimizer doesn't delete these!
		 * 
		 * may not be necessary on PC
		 */

		read_dump_0 = PLX_GETL(PLX_EEPROM_USER);
		read_dump_1 = IHCP_GETL(IHCP_MODE);

		/*
		 * clear sleeping flag so sleeping process will know we got a normal interrupt
		 * set wake up needed flag
		 */

		unit_p->sleeping = 0;
		wake_up_needed = 1;

	}

	/*
	 * release mutex
	 */

	mutex_unlock(&unit_p->mutex);

	if (wake_up_needed) {

		/*
		 * wake up sleeping process
		 */

		DPRINT((KERN_DEBUG
			"ihcp_intr: instance %d: waking up sleeping process\n",
			instance));

		cond_broadcast(&unit_p->cond);

	}

	return;

}




/*
 * strategy expects the virtual address of the user buffer (portion) to be output
 * at this point, it must be quad byte aligned and a byte count that must be 0 mod 4
 * 
 * if we are in user buffer DMA mode,
 * strategy will kiomap the buffer (pin it in physical memory), construct an sg list of buffer pages,
 * map the pages to the bus (using the dynamic mapping stuff), build the dma scatter-gather list in
 * iopb memory, synch the iopb memory, and do a single dma block output
 *
 * if we are in copy user buffer mode,
 * strategy will copy the to kernel memory buffer, and build a "fake" kiobuf
 * 
 * if the count parameter is zero, strategy will wait for fifo less than 1/2 full, but not do any output
 * 
 * THERE ARE SOME ERROR AND OTHER RETURNS IN THIS CODE -  IF WE EVER NEED A MUTEX, GATHER THEM WITH GOTOs, SO
 * WE WON'T ESCAPE OUT THE SIDE OF THE MUTEX -- BE CAREFUL TO UNMAP THINGS ON ERROR (IF ANYTHING WAS MAPPED
 * AT ERROR TIME)
 */


static int ihcp_strategy(struct ihcp_unit_t *unit_p, const char *dma_buf_p,
			 int dma_count)
{

	int		instance;	/* our instance (minor) number          */
	u_long		temp;		/* you guessed it                       */
	int		wait_return;	/* saved value returned from con_wait   */
	int		num_iopbs;	/* count of iopbs used & final sg size  */
	int		iopb_number;	/* iopb index - for readability         */
	int		kio_pages = 0;	/* count of user pages mapped           */
	int		kio_offset;	/* offset of data in mapped pages       */
	int		kio_length;	/* length of data in mapped pages       */
	struct page	**maplist;	/* pointer to kiobuf page pointers	*/
	uint32_t	temp_bus_addr;	/*                                      */
	int		temp_length;	/*                                      */
	uint32_t	next_iopb_add;	/* bus add of next iopb                 */
	int		error;		/* error code for error returns         */
	int		i;		/* the prototypical loop variable       */

	/*
	 * get instance number from unit structure
	 */

	instance = unit_p->instance;

	DPRINT((KERN_DEBUG
		"ihcp_strategy: instance %d: entering strategy, unit_p = 0x%p, buf_p = 0x%p, count = 0x%x\n",
		instance, unit_p, dma_buf_p, dma_count));

	/*
	 * make sure buffer is quad byte aligned and count is zero mod four
	 * this is not strictly necessary, but makes for a more generic strategy routine
	 * and check for max allowable size
	 */

	if ((u_long) dma_buf_p & 0x3) {
		printk(KERN_ERR
		       "ihcp_strategy: instance %d: dma buffer not quad word aligned!\n",
		       instance);
		return -EINVAL;
	}

	if (dma_count & 0x3) {
		printk(KERN_ERR
		       "ihcp_strategy: instance %d: dma count not 0 mod 4!\n",
		       instance);
		return -EINVAL;
	}

	if (dma_count > ihcp_max_buf_bytes) {
		printk(KERN_ERR
		       "ihcp_strategy: instance %d: dma count too large!\n",
		       instance);
		return -EINVAL;
	}
	
	/*
	 * set flags to indicate dma wait - even if we don't wait, just so flags will reflect
	 * that we were in strategy
	 * error flags are not cleared here so they will survive
	 * multiple calls to strategy by write
	 */

	unit_p->unit_flags |= IHCP_DMA_WAIT;

	/*
	 * we should now have an aligned buffer, with 0 or more bytes in it
	 * if total buffer size (remaining) is less than 4, skip dma stuff,
	 * and wait for <half full, even if byte count is 0
	 * 
	 * zero the iopb count here - it will be used later when unmapping (if necessary) and will tell
	 * us if we need to do any unmapping
	 */

	num_iopbs = 0;

	if (dma_count > 3) {

		DPRINT((KERN_DEBUG
			"ihcp_strategy: instance %d: dma required, buf virt add = 0x%p, size = 0x%x\n",
			instance, dma_buf_p, dma_count));

		if(unit_p->dma_copy_buf_p) {

			/*
			 * we are in copy mode
			 * copy user buffer to dma buffer and set up kiobuf
			 * we don't need to map_user_kiobuf since already in kernel
			 */

			DPRINT((KERN_DEBUG "ihcp_strategy: instance %d: copying user buffer to kernel buffer\n",
					instance));

			if (copy_from_user(unit_p->dma_copy_buf_p, dma_buf_p, dma_count)) {
				printk(KERN_ERR
					"ihcp_strategy: instance %d: copy_from_user error!\n",
					instance);
				return -EIO;
			}

			unit_p->kiobuf_p->offset = 0;
			unit_p->kiobuf_p->length = dma_count;
			unit_p->kiobuf_p->nr_pages = (dma_count + PAGE_SIZE -1)/PAGE_SIZE;

			DPRINT((KERN_DEBUG
					"ihcp_strategy: instance %d: copy_from_user successful\n",
					instance));
		} else {
			
			/*
			 * we are in user buffer DMA mode
			 * map the user buffer to kernel space, and pin it using kiomap
		 	 */

			DPRINT((KERN_DEBUG "ihcp_strategy: instance %d: mapping and pinning user buffer\n",
					instance));

			if (
			    (error =
		     		map_user_kiobuf(WRITE, unit_p->kiobuf_p,
					     (unsigned long) dma_buf_p,
				    	 dma_count))) {
				printk(KERN_ERR
				       "ihcp_strategy: instance %d: map_user_kiobuf error!\n",
				       instance);
				return error;
			}


			DPRINT((KERN_DEBUG
					"ihcp_strategy: instance %d: map_user_kiobuf successful\n",
					instance));
		}

		/*
		 * pull out some kiobuf elements, for easy reading (& typing)
		 * we already know some of these, and may not use others
		 */

		kio_pages = unit_p->kiobuf_p->nr_pages;
		kio_offset = unit_p->kiobuf_p->offset;
		kio_length = unit_p->kiobuf_p->length;

		DPRINT((KERN_DEBUG
			"ihcp_strategy: instance %d: kio_pages = 0x%x, kio_offset = 0x%x, kio_length = 0x%x\n",
			instance, kio_pages, kio_offset, kio_length));

		/*
		 * do a sanity check on the number of pages - it better not exceed the sg list max size
		 */

		if (kio_pages > ihcp_max_sg_length) {
			printk(KERN_ERR
			       "ihcp_strategy: instance %d: too many kiobuf pages!\n",
			       instance);
			if (!unit_p->dma_copy_buf_p) unmap_kiobuf(unit_p->kiobuf_p);
			return -EIO;
		}

		/*
		 * get pointer to array of page struct pointers
		 */

		maplist = unit_p->kiobuf_p->maplist;

		/*
		 * grind through the kio map list, extracting addresses and lengths for the sg list
		 * don't forget the initial offset, and total length
		 */

		for (i = 0; i < kio_pages; i++) {
			
			temp_length =
			    (PAGE_SIZE - kio_offset) <
			    kio_length ? PAGE_SIZE -
			    kio_offset : kio_length;

			/*
			 * use macro - .address or .page/.offset version of sg list selected
			 * by compile time option
			 */

			IHCP_SG_LINK(unit_p->sg_list_p[i], maplist[i], kio_offset, temp_length);

			kio_length -= temp_length;
			kio_offset = 0;
		}

		/*
		 * map the sg list (which may combine contiguous entries
		 * there is no error return, as I am told that the mapping resources are huge
		 */

		num_iopbs =
		    pci_map_sg(unit_p->pci_dev_p, unit_p->sg_list_p,
			       kio_pages, PCI_DMA_TODEVICE);

		DPRINT((KERN_DEBUG
			"ihcp_strategy: instance %d: scatter list mapped, num_iopbs = 0x%x\n",
			instance, num_iopbs));

		/*
		 * make sure num_iopbs <= kio_pages (we should probably just trust the mapping code
		 */

		if (num_iopbs > kio_pages) {

			printk(KERN_ERR
			       "ihcp_strategy: instance %d: num_iopbs > kio_pages!\n",
			       instance);

			pci_unmap_sg(unit_p->pci_dev_p, unit_p->sg_list_p,
				     kio_pages, PCI_DMA_TODEVICE);
			if (!unit_p->dma_copy_buf_p) unmap_kiobuf(unit_p->kiobuf_p);
			return -EIO;

		}

		/*
		 * now assemble the scatter list into a form that the PLX chip wants to see, and build iopb list
		 * use IOPB macros to ease possible later mods to swizzle byte lanes
		 */

		for (iopb_number = 0; iopb_number < num_iopbs;
		     iopb_number++) {

			temp_bus_addr =
			    (uint32_t)
			    sg_dma_address(&
					   (unit_p->
					    sg_list_p[iopb_number]));
			temp_length =
			    sg_dma_len(&(unit_p->sg_list_p[iopb_number]));

			IOPB_PUTL((iopb_number * IOPB_SIZE) +
				  IOPB_PCI_ADDRESS, temp_bus_addr);	/* buf segment bus address      */
			IOPB_PUTL((iopb_number * IOPB_SIZE) +
				  IOPB_TRANSFER_SIZE, temp_length);	/* buf segment size             */
			IOPB_PUTL((iopb_number * IOPB_SIZE) +
				  IOPB_LOCAL_ADDRESS, IHCP_32_BIT_DATA_OUT);	/* data destination             */

			DPRINT((KERN_DEBUG
				"ihcp_strategy: instance %d: processing iopb number 0x%x, buf seg addr = 0x%x, buf seg length = 0x%x\n",
				instance, iopb_number, temp_bus_addr,
				temp_length));

			/*
			 * the next descriptor pointer element of the iopb includes the location bit
			 * (pci or local memory), the end of chain bit, the interrupt after terminal
			 * count bit (not used), the direction of transfer bit (0 = output to device),
			 * and the bus add of the next iopb
			 * 
			 * if the iopb_number = num_iopbs -1, this is the last one so flag it,
			 * and don't set a next iopb address
			 */

			if (iopb_number < (num_iopbs - 1)) {

				next_iopb_add =
				    DMA_NEXT_ADDRESS_MASK &
				    ((uint32_t) unit_p->dma_handle +
						((iopb_number + 1) *
						 IOPB_SIZE));

				DPRINT((KERN_DEBUG
					"ihcp_strategy: instance %d: not last iopb, next iopb add = 0x%x\n",
					instance, next_iopb_add));
				IOPB_PUTL((iopb_number * IOPB_SIZE) +
					  IOPB_NEXT_IOPB,
					  DMA_CHAIN_IN_PCI_MEM |
					  next_iopb_add);

			} else {

				DPRINT((KERN_DEBUG
					"ihcp_strategy: instance %d: last iopb, setting end of chain bit\n",
					instance));
				IOPB_PUTL((iopb_number * IOPB_SIZE) +
					  IOPB_NEXT_IOPB,
					  DMA_CHAIN_IN_PCI_MEM |
					  DMA_END_OF_CHAIN);

			}

		}		/* end of for iopb_number       */

		/*
		 * synch iopb memory for dma - NOP on a pc- necessary on other
		 * architectures to ensure coherence betweene list and device
		 * synch only as much iopb space as was touched this time
		 */

		pci_dma_sync_single(unit_p->pci_dev_p, unit_p->dma_handle,
				    num_iopbs * IOPB_SIZE,
				    PCI_DMA_TODEVICE);

		DPRINT((KERN_DEBUG
			"ihcp_strategy: instance %d: iopb memory synched for DMA\n",
			instance));

		/*
		 * program plx dma logic on board - in chaining mode
		 * buffer is already aligned, and count masked
		 * 
		 * make sure that any left over dma done interrupt
		 * is cleared
		 */

		DPRINT((KERN_DEBUG
			"ihcp_strategy: instance %d: programming PLX chip for DMA\n",
			instance));

		PLX_PUTL(PLX_DMA_CMD_STAT_BOTH, DMA_CLEAR_INTERRUPT);

		PLX_PUTL(PLX_DMA_MODE_0,
			  DMA_BUS_32_BIT | DMA_WAIT_1 | DMA_BURST_ENABLE |
			  DMA_CHAIN_ENABLE | DMA_DONE_INTERRUPT_ENB |
			  DMA_LOCAL_ADD_HOLD | DMA_DEMAND_MODE);				/* mode register                                */
		PLX_PUTL(PLX_DMA_DESC_PTR_0, DMA_CHAIN_IN_PCI_MEM | unit_p->dma_handle);	/* outputl, list in pci mem, list pointer       */
		PLX_PUTL(PLX_DMA_CMD_STAT_BOTH, DMA_ENABLE);						/* start DMA in two steps - PLX bug             */
		PLX_PUTL(PLX_DMA_CMD_STAT_BOTH, DMA_ENABLE | DMA_START);				/* actually fire up xfer                        */

		/*
		 * the 9060 drives the local interrupt out true when a dma
		 * interrupt happens
		 * 
		 * we have to loop that output back to the local interrupt input
		 * external to the chip to get a dma interrupt to the pci bus
		 */

		/*
		 * save processor flags, disable interrupts, and take spinlock w/mutex
		 * BEFORE ENABLING BOARD INTERRUPTS!
		 * 
		 * THERE ARE TWO PLACES IN STRATEGY WHERE THIS MAY BE DONE, BUT ONLY ONE PLACE WHERE FLAGS ARE RESTORED
		 */

		DPRINT((KERN_DEBUG
			"ihcp_strategy: instance %d: disabling cpu ints and enabling board dma done int\n",
			instance));

		mutex_lock(&unit_p->mutex);

		PLX_PUTL(PLX_INT_CSTAT, PCI_INTERRUPT_ENABLE |
			 PCI_LOCAL_INT_ENABLE | LOCAL_INT_OUT_ENABLE |
			 LOCAL_DMA_0_INT_ENABLE);

	} else {

		DPRINT((KERN_DEBUG
			"ihcp_strategy: instance %d: no dma, waiting for <1/2 full\n",
			instance));

		/*
		 * we need to wait for fifo less than half full
		 * it may already be less than half full - if so
		 * return good status
		 */

		if (IHCP_GETL(IHCP_INTERFACE_STATUS) & FIFO_NOT_HALF_FULL)
			return 0;

		DPRINT((KERN_DEBUG
			"ihcp_strategy: instance %d: not <1/2 full, setting up for interrupt\n",
			instance));

		/*
		 * clear any left over interrupt flag, then
		 * enable plx and ihcp interrupt for fifo less than half full
		 * we don't or into ihcp or plx mask reg since we know exactly which ints to enb
		 */

		IHCP_PUTL(IHCP_INTERFACE_CONTROL, CLEAR_INTERRUPT_FLAG);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

		/*
		 * save processor flags, disable interrupts, and take spinlock w/mutex
		 * BEFORE ENABLING BOARD INTERRUPTS!
		 * 
		 * THERE ARE TWO PLACES IN STRATEGY WHERE THIS MAY BE DONE, BUT ONLY ONE PLACE WHERE FLAGS ARE RESTORED
		 */

		DPRINT((KERN_DEBUG
			"ihcp_strategy: instance %d: disabling processor ints, and enabling board fifo < 1/2 full int\n",
			instance));

		mutex_lock(&unit_p->mutex);

		IHCP_PUTL(IHCP_INTERRUPT_MASK, FIFO_NOT_HALF_INT_ENB);
		PLX_PUTL(PLX_INT_CSTAT,
			 (PCI_INTERRUPT_ENABLE | PCI_LOCAL_INT_ENABLE));

	}

	/*
	 * we arrive here after setting up for a dma wait, or a less than half full wait
	 * 
	 * set sleeping flag - so we can detect abnormal sleep termination later ( int code will clear sleeping flag)
	 * 
	 * board interrupts are already enabled, and processor interrupts disabled, to avoid race conditions - sleep will re-enable
	 * 
	 * snooze until done or <half full interrupt, timeout, or signal
	 */

	unit_p->sleeping = 1;

	/*
	 * sleep until done, timeout, or signal
	 * 
	 * kernel source looks like this returns 0 for timeout, >0 for non-timeout
	 */

	DPRINT((KERN_DEBUG
		"ihcp_strategy: instance %d: calling cond_timed_wait_rel, jiffies = %ld, unit_p->dma_time = %ld\n",
		instance, jiffies, unit_p->dma_time));

	wait_return =
	    cond_timed_wait_rel(&unit_p->cond, &unit_p->mutex,
				unit_p->dma_time);

	/*
	 * make sure interrupt enables are off and pause dma (which will normally
	 * already be done)
	 */

	PLX_PUTL(PLX_DMA_CMD_STAT_BOTH, 0);
	IHCP_PUTL(IHCP_INTERRUPT_MASK, 0);
	PLX_PUTL(PLX_INT_CSTAT, 0);

	/*
	 * re-enable interrupts (if previously on)
	 * and release spin lock via mutex
	 */

	mutex_unlock(&unit_p->mutex);

	/*
	 * if we were doing dma, and
	 * if dma is not done (after sig or timeout), force an abort
	 * and issue a master clear to flush ihcp buffers
	 * 
	 * abort may take some time to complete, so use delay and
	 * register read to give it some time
	 * 
	 * we have to be sure we were actually doing dma - since we need
	 * to unmap the sg list and kiobuf
	 */

	if (num_iopbs && !(PLX_GETL(PLX_DMA_CMD_STAT_BOTH) & DMA_DONE)) {

		DPRINT((KERN_DEBUG
			"ihcp_strategy: instance %d: dma not done after cond_timed_wait_rel() return\n",
			instance));

		PLX_PUTL(PLX_DMA_CMD_STAT_BOTH, DMA_ABORT);

		/*
		 * get mode - so we can restore it after a mclr
		 * 
		 * mclr clears the fifos, and also makes them assert dma rq
		 * but...mclr holds off dma rq while mclr is asserted, and since the 9060
		 * likes a synchronous dma rq, we will wait a while before hitting mclr
		 * 
		 * 10us of delay seems like a lot, but we will only be in this code
		 * following an error
		 * 
		 * it is not clear whether the 9060 needs dma rq in order to 
		 * finish an abort
		 * 
		 * if it does, it may transfer a word or two after the mclr,
		 * so we will repeat the mclr later
		 */

		udelay(5);

		temp = IHCP_GETL(IHCP_MODE);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, MASTER_CLEAR);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

		/*
		 * force write cache to flush - so delay appears on the bus
		 */

		read_dump_0 = IHCP_GETL(IHCP_INTERRUPT_MASK);

		udelay(5);

		IHCP_PUTL(IHCP_INTERFACE_CONTROL, MASTER_CLEAR);
		IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);
		IHCP_PUTL(IHCP_MODE, temp);

		/*
		 * make sure dma actually says done
		 */

		if (!(PLX_GETL(PLX_DMA_CMD_STAT_BOTH) & DMA_DONE)) {
			printk(KERN_ERR
			       "ihcp_strategy: instance %d: dma not done after abort!\n",
			       instance);
			pci_unmap_sg(unit_p->pci_dev_p, unit_p->sg_list_p,
				     kio_pages, PCI_DMA_TODEVICE);
			if (!unit_p->dma_copy_buf_p) unmap_kiobuf(unit_p->kiobuf_p);
			return -EIO;
		}
	}

	/*
	 * now we can safely make sure that the done interrupt is clear,
	 * and reset the ihcp interrupt flag
	 * 
	 * (the above is probably not necessary after normal dma done)
	 */

	PLX_PUTL(PLX_DMA_CMD_STAT_BOTH, DMA_CLEAR_INTERRUPT);
	IHCP_PUTL(IHCP_INTERFACE_CONTROL, CLEAR_INTERRUPT_FLAG);
	IHCP_PUTL(IHCP_INTERFACE_CONTROL, 0);

	/*
	 * if dma was performed, unmap the sg list (NOP on pc, release iommu mappings on other machines)
	 * unmap sg wants the original sg list length
	 */

	if (num_iopbs) {

		DPRINT((KERN_DEBUG
			"ihcp_strategy: instance %d: unmapping sg list and kiobuf (if mapped) after dma\n",
			instance));

		pci_unmap_sg(unit_p->pci_dev_p, unit_p->sg_list_p,
			     kio_pages, PCI_DMA_TODEVICE);
		if (!unit_p->dma_copy_buf_p) unmap_kiobuf(unit_p->kiobuf_p);

	}

	/*
	 * find out why we woke up
	 */

	if (unit_p->sleeping) {

		unit_p->sleeping = 0;

		DPRINT((KERN_DEBUG
			"ihcp_strategy: instance %d: abnormal return from cond_timed_wait_rel()!\n",
			instance));

		/*
		 * check for timeout
		 */

		if (wait_return == COND_WAIT_TIMEOUT) {

			/*
			 * timeout is always an error so print message and return error code to write
			 * and flag timeout in unit flags
			 */

			printk(KERN_ERR
			       "ihcp_strategy: instance %d: timeout while waiting for interrupt!\n",
			       instance);
			unit_p->unit_flags |= IHCP_DMA_TIMEOUT;
			return -EIO;
		}

		/*
		 * check for signal if not masked (blocked is a signal bitmask)
		 */

		if (wait_return == COND_WAIT_SIGNAL) {

			/*
			 * flag signal received and print error message
			 */

			printk(KERN_WARNING
			       "ihcp_strategy: instance %d: signal received while waiting for interrupt\n",
			       instance);
			unit_p->unit_flags |= IHCP_SIG_RECEIVED;
			return -EINTR;
		}

	}


	/* end of if sleeping       */
	/*
	 * if we get here - just return good status
	 * 
	 * there is no way to figure a residual count with the hardware at hand
	 * so we don't even try
	 */

	return 0;

}

/*
 * mutex and conditional variable routines - were in ikon_mutex.c
 * Kaz Kylheku & wdw
 */

void mutex_init(mutex_t * mx)
{

	spin_lock_init(&mx->spin);

}

void mutex_lock(mutex_t * mx)
{

	unsigned long flags;

	spin_lock_irqsave(&mx->spin, flags);
	mx->flags = flags;

}

void mutex_unlock(mutex_t * mx)
{

	unsigned long flags = mx->flags;
	spin_unlock_irqrestore(&mx->spin, flags);

}

void cond_init(cond_t * cv)
{

	init_waitqueue_head(&cv->queue);

}

int cond_timed_wait_rel(cond_t * cv, mutex_t * mx, long jiff_delta)
{

	wait_queue_t wait;
	int ret;
	long remaining;

	init_waitqueue_entry(&wait, current);
	add_wait_queue(&cv->queue, &wait);
	current->state = TASK_INTERRUPTIBLE;

	mutex_unlock(mx);

	remaining = schedule_timeout(jiff_delta);
	ret =
	    remaining ? signal_pending(current) ? COND_WAIT_SIGNAL :
	    COND_WAIT_SUCCESS : COND_WAIT_TIMEOUT;
	current->state = TASK_RUNNING;
	remove_wait_queue(&cv->queue, &wait);

	mutex_lock(mx);

	return ret;

}

void cond_broadcast(cond_t * cv)
{

	wake_up_interruptible(&cv->queue);

}
